"""
This module contains functions useful for querying the database.
These functions are also available in the `chemrecon` scope.
"""
from chemrecon.query.find_entry import (
    find_entry,
    find_compound_entry, find_reaction_entry, find_enzyme_entry,
    find_structure_representation_entry, find_aam_representation_entry,
    find_structure_entry, find_aam_entry
)

# Relation getters
from chemrecon.query.get_relations import (
    get_relations_from_entry,
    get_all_relations,
)

# from chemrecon.query.create_entry import (
#     entry,
#     compound_entry, reaction_entry, enzyme_entry,
#     aam_representation_entry, structure_representation_entry,
#     structure_entry, aam_entry,
#     enzyme_from_ec_number,
#     entry_from_identifiers_org
# )




from __future__ import annotations

from chemrecon.schema.procedural_relation_entrygraph import ProceduralRelationEG
from chemrecon.schema.direction import Direction
from chemrecon.entrygraph.scoring import Scorer
from chemrecon.schema.entry_types.compound import Compound
from chemrecon.schema.entry_types.molstructure import MolStructure
from chemrecon.schema.relation_types_composed.compound_has_molstructure_relation import CompoundHasMolStructure
from chemrecon.schema.relation_types_source.compound_reference_relation import CompoundReference
from chemrecon.schema.relation_types_source.molstructure_standardisation_relation import MolStructureStandardization
from chemrecon.entrygraph.explorationprotocol import ExplorationProtocol

ProtocolSelectStructure = ExplorationProtocol(
    relation_types = {
        (CompoundReference, Direction.SYMMETRIC),
        (CompoundHasMolStructure, Direction.FORWARDS),
        (MolStructureStandardization, Direction.FORWARDS),
    },
)
eg_scorer_compound_select_structure = Scorer[MolStructure](
    score_entry_type = MolStructure
)

class CompoundSelectStructure(ProceduralRelationEG[Compound, MolStructure]):
    """ Procedural relation which computes the most related molecular structures for a given compound using entry graph
        scoring.
    """

    # Database
    _table_name = 'CompoundSelectStructure'
    source_entrytype = Compound
    target_entrytype = MolStructure

    # For EG
    protocol = ProtocolSelectStructure
    entrygraph_scorer = eg_scorer_compound_select_structure



from __future__ import annotations

import re

import networkx as nx
from rdkit import Chem as rdk
from rdkit.Chem.rdChemReactions import ChemicalReaction

from chemrecon.chem.gml.gml import remove_residues, GMLGraph


atom_regex = re.compile(r'([A-Z][a-z]*)([0-9])?([+\-])?')


# Molecule
def gml_to_rdkit_mol(gml: str) -> rdk.Mol:
    gmlgraph = GMLGraph(gml)
    rdk_mol = rdk.EditableMol(rdk.Mol())
    node_index: dict[int, int] = {}

    # Atoms
    for nx_index, nx_properties in gmlgraph.graph.nodes.items():
        # Add atom with element (charge?) given by the label, and atom map number given by the index
        atom = None
        label: str = nx_properties['label']
        if label == '*':
            atom = atom = rdk.AtomFromSmiles(f'[*]')
        else:
            match = atom_regex.match(label)
            charge = 0
            try:
                symbol = match.group(1)
                plusminus = match.group(3)
                if plusminus is None:
                    charge = 0
                else:
                    if plusminus == '+':
                        charge = 1
                    elif plusminus == '-':
                        charge = -1
                    else:
                        raise Exception()

                chargeabs = match.group(2)
                if chargeabs:
                    charge * int(chargeabs)

            except Exception:
                raise ValueError(f'Cannot convert S_GML: Atom label: {label}')
                # error.

            if symbol == 'R':
                # Radical
                # TODO
                atom = rdk.AtomFromSmiles(f'[*]')
            elif symbol.startswith('Alias'):
                # 'Alias in GML?'
                # TODO what does this mean exactly
                atom = rdk.AtomFromSmiles(f'[*]')
            else:
                atom = rdk.AtomFromSmiles(f'[{symbol}]')

            # Set charge based on match
            atom.SetFormalCharge(charge)

        # Atom is set
        if atom is None:
            raise ValueError(f'Cannot convert S_GML: Atom label: {label}')

        atom.SetAtomMapNum(nx_index)
        atom_index = rdk_mol.AddAtom(atom)
        node_index[nx_index] = atom_index

    # Edges
    for a, b in gmlgraph.graph.edges():
        index_a = node_index[a]
        index_b = node_index[b]
        label: str = gmlgraph.graph.get_edge_data(a, b)['label']
        match label:
            case '-':
                bondtype = rdk.BondType.SINGLE
            case '=':
                bondtype = rdk.BondType.DOUBLE
            case '#':
                bondtype = rdk.BondType.TRIPLE
            case ':':
                bondtype = rdk.BondType.AROMATIC
            case _:
                raise ValueError(f'Unknown bond type: {label}.')
        rdk_mol.AddBond(index_a, index_b, bondtype)

    # Mol is done, change to uneditable
    rdk_mol_final = rdk_mol.GetMol()
    return rdk_mol_final

# Reaction
def gml_to_rdkit_reaction(gml: str) -> ChemicalReaction:
    """ Relies on S_GML in the format outputted by the M-CSA script.
    """

    # TODO change to re-use above code?

    # Load two S_GML strings
    l_str = ''
    r_str = ''
    step: int = 0
    currentstr: str = ''
    for l in gml.splitlines():
        if l.startswith('graph') or l.startswith('left') or l.startswith('right'):
            if step == 1:
                l_str = currentstr
            elif step == 2:
                r_str = currentstr
            # Start new string
            step += 1
            currentstr = ''
        # Add to current string
        currentstr += l

    assert step == 2
    r_str = currentstr

    # Get the components and turn into MolInstances
        # TODO should not be necessary
    l_chemgraph = remove_residues(GMLGraph(l_str.replace('left', 'graph')))
    r_chemgraph = remove_residues(GMLGraph(r_str.replace('right', 'graph')))

    # For each component, add it
    rdk_r = ChemicalReaction()
    rdk_reaction = ChemicalReaction()

    for chemgraph, side in [[l_chemgraph, -1], [r_chemgraph, 1]]:
        for component in chemgraph.get_components_subgraphs():
            component: nx.Graph
            rdk_mol = rdk.EditableMol(rdk.Mol())
            node_index: dict[int, int] = {}

            # Atoms
            for nx_index, nx_properties in component.nodes.items():
                # Add atom with element (charge?) given by the label, and atom map number given by the index
                atom = None
                label: str = nx_properties['label']
                if label == '*':
                    atom = rdk.AtomFromSmiles(f'[*]')
                else:
                    match = atom_regex.match(label)
                    charge = 0
                    try:
                        symbol = match.group(1)
                        plusminus = match.group(3)
                        if plusminus is None:
                            charge = 0
                        else:
                            if plusminus == '+':
                                charge = 1
                            elif plusminus == '-':
                                charge = -1
                            else:
                                raise Exception()

                        chargeabs = match.group(2)
                        if chargeabs:
                            charge * int(chargeabs)

                    except Exception:
                        raise ValueError(f'Cannot convert S_GML: Atom label: {label}')
                        # error.
                    if symbol == 'R':
                        # Radical
                        # TODO
                        atom = rdk.AtomFromSmiles(f'[*]')
                    elif symbol.startswith('Alias'):
                        # 'Alias in GML?'
                        # TODO what does this mean exactly
                        atom = rdk.AtomFromSmiles(f'[*]')
                    else:
                        atom = rdk.AtomFromSmiles(f'[{symbol}]')


                if atom is None:
                    raise ValueError(f'Cannot convert S_GML: Atom label: {label}')
                atom.SetFormalCharge(charge)
                atom.SetAtomMapNum(nx_index)
                atom_index = rdk_mol.AddAtom(atom)
                node_index[nx_index] = atom_index

            # Edges
            for a, b in component.edges():
                index_a = node_index[a]
                index_b = node_index[b]
                label: str = component.get_edge_data(a, b)['label']
                match label:
                    case '-':
                        bondtype = rdk.BondType.SINGLE
                    case '=':
                        bondtype = rdk.BondType.DOUBLE
                    case '#':
                        bondtype = rdk.BondType.TRIPLE
                    case ':':
                        bondtype = rdk.BondType.AROMATIC
                    case _:
                        raise ValueError(f'Unknown bond type: {label}.')
                rdk_mol.AddBond(index_a, index_b, bondtype)

                pass

            # Mol is done, change to uneditable
            rdk_mol_final = rdk_mol.GetMol()
            # TODO ADD mol to reaction
            match side:
                case -1:
                    rdk_reaction.AddReactantTemplate(rdk_mol_final)
                case 1:
                    rdk_reaction.AddProductTemplate(rdk_mol_final)

    # Reaction now completed
    return rdk_reaction



# Match the identifiers org prefix as group 1, and the actual id as group 2
import re
from typing import Optional

from chemrecon.core.id_types import (IdentifierType, IdentifierTypeAAM, IdentifierTypeCompound,
                           IdentifierTypeEnzyme, IdentifierTypeReaction,
                           IdentifierTypeStructureRepresentation,
                           id_types_aam, id_types_compound,
                           id_types_enzyme, id_types_reaction, id_types_structure_representation,
                           identifiers_org_dict, )

identifiers_org_regex = re.compile(r'(https?://identifiers.org/[a-zA-Z./\-]+/)(.+)')

# Sub-lookup dicts for identifier type names for each entrytype (compounds, reactions, etc)
id_type_name_lookup_subsdicts: dict[type[IdentifierType], dict[str, IdentifierType]] = dict()
typelists: dict[type[IdentifierType], list[IdentifierType]] = {
    IdentifierTypeCompound: id_types_compound,
    IdentifierTypeStructureRepresentation: id_types_structure_representation,
    IdentifierTypeReaction: id_types_reaction,
    IdentifierTypeEnzyme: id_types_enzyme,
    IdentifierTypeAAM: id_types_aam
}
for entrytype_idtype, idtypes in typelists.items():
    subdict: dict[str, IdentifierType] = dict()
    for idtype in idtypes:
        subdict[idtype.name.lower()] = idtype
        subdict[idtype.shortname.lower()] = idtype
        for name in idtype.alt_names:
            subdict[name.lower()] = idtype
    id_type_name_lookup_subsdicts[entrytype_idtype] = subdict



def get_id_type[T: IdentifierType](
    type_str: str,
    identifier_supertype: type[T]
) -> Optional[T]:
    """ Get an identifier type by a strign description. """
    try:
        return id_type_name_lookup_subsdicts[identifier_supertype][type_str.lower()]
    except KeyError:
        return None

def get_id_type_from_name[T: IdentifierType](name: str) -> T:
    """ Get an identifier type (of type T, e.g. Compound, Reaction, Struct representation, ...) """
    raise NotImplementedError

def get_possible_id_types_from_source_id[T: IdentifierType](name: str) -> list[T]:
    """ Get a list of possible compatible identifiers (of type T, e.g. Compound, Reaction,
    Struct representation, ...) """
    raise NotImplementedError

def get_id_from_identifiers_org(
    string: str,
    identifier_supertype: type[IdentifierType]
) -> Optional[tuple[IdentifierType, str]]:
    """ Get an identifier type (of type T, e.g. Compound, Reaction, Struct representation, ...) """
    match = re.match(identifiers_org_regex, string)
    if match:
        prefix = match.group(1)
        match_id = match.group(2).removesuffix('/')     # TODO should trailing '/' be removed in all cases?
        try:
            id_type: IdentifierType = identifiers_org_dict[prefix]
            if not isinstance(id_type, identifier_supertype):
                return None
            return (id_type, match_id)
        except KeyError:
            return None
    else:
        return None



""" Defines a wrapper class for RDKit molecules.
    A Mol may either be a MolTemplate or a MolInstance.
    - MolInstance: The molecule as it appears in a reaction concretely.
    - MolTemplate: More abstract representation of the molecule in general.
"""
from __future__ import annotations

import ast
from copy import copy, deepcopy
from typing import Type, Optional, Any

import rdkit.Chem as rdk
from rdkit.Chem import rdFingerprintGenerator as rdk_fp
import rdkit.Chem.MolStandardize.rdMolStandardize as rdk_std
from rdkit import DataStructs as rdk_ds

from chemrecon.chem.elements import atomicnum_element
from chemrecon.chem.sumformula import SumFormula
from chemrecon.schema import MolStructure
from chemrecon.schema.enums import FeatureEnum

# RDkit generators
fpgen = rdk_fp.GetRDKitFPGenerator()

# Main class
# --------------------------------------------------------------------------------------------------------------
class Mol:
    """ Wrapper for RDKit Mol
    """
    # TODO rename/refactor to 'MolStructure'?
    mol:                rdk.Mol

    smiles:             Optional[str]
    features:           set[Type[Feature]]  # Features for which the rdk_mol is standardized
    n_atoms:            int
    mass:               float
    charge:             int
    molformula:         SumFormula
    provenance:         Optional[str]
    propdict:           dict[str, str]  # Properties



    # Optional fingerprint
    fp:                 Optional[Any]

    def __init__(
        self,
        rdk_mol: rdk.Mol,
        set_features: set[Type[Feature]] = None,
        provenance: Optional[str] = None
    ):
        if rdk_mol is None:
            raise ValueError('Cannot create Mol from None.')

        # Normalize the molecule
        try:
            self.mol = rdk_std.Normalize(rdk_mol)       # TODO what exactly does normalisation do?
        except ValueError as e:
            # Could not create molecule, or molecule is None?
            print(e)
            self.mol = None
            self.smiles = None
            # return
            raise ValueError(f'Could not create molecule: {e}')

        # Set smiles for printing
        self.smiles = rdk.MolToSmiles(self.mol)

        # Fix SMILES bug (non-C lowercase atom symbols)
        # https://github.com/rdkit/rdkit/issues/3697
        self.smiles = self.smiles.replace('[o', '[O')

        if self.smiles is None:
            raise ValueError('TODO fix')

        # Compute featuers
        if set_features is not None:
            # Features set in constructor
            self.features = set_features
        else:
            self.features = set()
            for f in feats:
                # Check standardization and set
                if f.is_standardized(mol = self):
                    self.features.add(f)

        # Get properties from name if initialized through an RXN file
        try:
            molprops_rxn = rdk_mol.GetProp('_Name')
            try:
                rxn_propdict = ast.literal_eval(molprops_rxn)
                self.propdict = rxn_propdict
            except Exception as e:
                self.propdict = dict()
        except KeyError:
            # Not applicable, skip
            self.propdict = dict()
            pass

        # Populate misc fields
        self.n_atoms = self.mol.GetNumAtoms(onlyExplicit = False)
        self.provenance = provenance

        # Get MolFormula
        self.molformula = self.get_molformula()

        # TODO n_atoms_tracked
        # TODO mass (simple)

        self.charge = rdk.GetFormalCharge(self.mol)

        # Set fp
        self.fp = None

        if self.provenance is None:
            # TODO fix
            pass

    # Converters
    # ----------------------------------------------------------------------------------------------------------
    def to_smiles(self) -> str:
        # TODO return more details
        return self.smiles

    # Database interfacing
    # ----------------------------------------------------------------------------------------------------------
    def to_database_struct(self) -> MolStructure:
        """ Create a database row to be inserted."""
        if self.mol is None and self.smiles is None:
            raise ValueError('Cannot convert None entry')

        # TODO fixes rdkit bug?
        # Fix SMILES bug (non-C lowercase atom symbols)
        # https://github.com/rdkit/rdkit/issues/3697
        self.smiles.replace(' [o', '[O')

        # Else, convert
        return MolStructure(
            smiles = self.to_smiles(),
            std_feats = [
                f.feature_enum for f in self.features
            ]
        )

        # TODO insert calculated properties

    # Calculate properties
    # ----------------------------------------------------------------------------------------------------------
    # TODO

    # Standardisation
    def get_standardised(self, feat_list: Optional[list[Feature]] = None) -> Mol:
        """ Get a new Mol standardised according to the given features. If None given, standardise according
            to all features.
        """
        feat_list = feat_list or [F, I, C, T, S]
        rdk_s: rdk.Mol = self.mol
        for f in feat_list:
            rdk_s = f.standardise_rdk(rdk_s)
        return Mol(rdk_s, set_features = feat_list, provenance = self.provenance)

    # Identity and similarity
    # ------------------------------------------------------------------------------------------------------------------
    def is_identical_up_to_map(self, other: Mol):
        """ Checks whether two Mols are identical except for extra details such as atom mapping numbers. """
        m: rdk.Mol = deepcopy(self.mol)
        m_: rdk.Mol = deepcopy(other.mol)

        for a in m.GetAtoms():
            a.SetAtomMapNum(0)
        for a in m_.GetAtoms():
            a.SetAtomMapNum(0)

        return rdk.MolToSmiles(m) == rdk.MolToSmiles(m_)

    def is_identical_up_to(self, other: Mol, features: set[Feature]) -> bool:
        # TODO
        raise NotImplementedError

    def generate_fingerprint(self):
        if self.fp is None:
            self.fp = fpgen.GetFingerprint(self.mol)

    def get_similarity(self, other: Mol) -> float:
        """ Get a similarity metric between two Mols. Should be 1 if the mols are identical,
            and approach zero as the difference increases.
        """
        self.generate_fingerprint()
        other.generate_fingerprint()
        return rdk_ds.TanimotoSimilarity(
            self.fp, other.fp
        )

    # Misc
    # ----------------------------------------------------------------------------------------------------------
    def feature_string(self) -> str:
        s = ''
        for f in feats:
            if f in self.features:
                s += f.symbol
            else:
                s += '-'
        return s

    def __repr__(self):
        return f'{self.smiles}    [{self.feature_string()}]'

    def __hash__(self):
        return f'{self.smiles}:::{self.feature_string()}'.__hash__()

    def __eq__(self, other: Mol):
        if isinstance(other, Mol):
            return self.smiles == other.smiles and self.features == other.features
        else:
            return False

    # Serialise
    def serialize(self) -> dict:
        return {
            'smiles': self.to_smiles(),
            'features': self.feature_string(),
            'molformula': 'TODO'  # TODO
        }

    # Get molecular formula
    def get_molformula(self) -> SumFormula:
        """ Returns the molecular formula
        """
        # TODO requires rdkit documentation
        formula_num: dict[int, int] = dict()
        for a in self.mol.GetAtoms():
            n: int = a.GetAtomicNum()
            formula_num[n] = formula_num.get(n, 0) + 1

        charge = rdk.GetFormalCharge(self.mol)

        # Translate to formula of elements and return
        return SumFormula(
            formula = {
                atomicnum_element[n]: i for n, i in formula_num.items()
            },
            charge = charge
        )

# Mol Instance and Template
# ----------------------------------------------------------------------------------------------------------------------
class MolInstance(Mol):
    """ MolInstance represents a particular instance of a molecule as it participates in a reaction.
    """

    n_atoms_tracked: int
    _atom_map: Optional[list[int]]  # For each atom in the MolInstance, gives the mapping.
    _atom_map_smiles_order: Optional[list[int]]

    def __init__(self, rdk_mol: rdk.Mol, set_features: set[Type[Feature]] = None, provenance: Optional[str] = None):
        super().__init__(rdk_mol, set_features = set_features, provenance = provenance)
        self._atom_map = None
        self._atom_map_smiles_order = None

    def to_mol_template(self) -> MolTemplate:
        """ Remove all information of the instance as it participates in a reaction (atom-to-atom map etc.)
            to generate a generic template of the structure.
        """
        mtemp = MolTemplate(
            rdk_mol = self.mol,
            set_features = self.features,
            provenance = self.provenance
        )

        # Carry over certain properties
        mtemp.propdict = self.propdict
        return mtemp

    def get_atom_map_in_native_order(self) -> list[int]:
        # The numbering of the atoms as in the reaction it is an instance of.
        if not self._atom_map:
            self._atom_map = [
                a.GetAtomMapNum()
                for a in self.mol.GetAtoms()
            ]

        return self._atom_map


    def get_atom_map_in_smiles_order(self, safe: bool = False) -> list[int]:
        # Same as getting the atom map, but with the ordering of the atoms in the smiles string
        # Needed to populate DB with atom-to-atom maps
        if not self._atom_map_smiles_order:
            smilesmol = rdk.MolFromSmiles(self.smiles, sanitize = not safe)
            self._atom_map_smiles_order = [
                a.GetAtomMapNum()
                for a in smilesmol.GetAtoms()
            ]

        return self._atom_map_smiles_order

    def serialize(self) -> dict:
        d = super().serialize()
        d.update({
            'template': self.to_mol_template().to_smiles()
        })
        return d


class MolTemplate(Mol):
    """ MolTemplate represents a consistent type of molecule, independent of its particular participation and
        mapping in a reaction.
    """
    pass

    def __init__(
        self,
        rdk_mol: rdk.Mol,
        set_features: set[Type[Feature]] = None,
        provenance: Optional[str] = None
    ):
        rdk_mol_ = deepcopy(rdk_mol)
        if not rdk_mol_:
            raise ValueError(f'Could not create molecule')

        # Remove atom-to-atom maps
        for a in rdk_mol_.GetAtoms():
            a.SetAtomMapNum(0)

        super().__init__(rdk_mol = rdk_mol_, set_features = set_features, provenance = provenance)

    def instantiate(self, mapping: list[int]) -> MolInstance:
        """ Instantiate a template with a map, to produce an instance of the template as it participates in a reaction.
        """
        rdk_mol_ = deepcopy(self.mol)
        for a, mapnum in zip(rdk_mol_.GetAtoms(), mapping):
            a.SetAtomMapNum(mapnum)
        return MolInstance(rdk_mol = rdk_mol_, set_features = self.features, provenance = 'from_template')


# Features and standardization
# ----------------------------------------------------------------------------------------------------------------------
# ABC
class Feature:
    symbol: str
    name: str
    feature_enum: FeatureEnum

    @classmethod
    def is_standardized(cls, mol: Mol) -> bool:
        raise NotImplementedError

    @classmethod
    def standardise(cls, mol: Mol) -> rdk.Mol:
        if mol.mol is None:
            raise ValueError
        return cls.standardise_rdk(mol.mol)

    @classmethod
    def standardise_rdk(cls, rdk_mol: rdk.Mol) -> rdk.Mol:
        raise NotImplementedError


# FICTS features
# --------------------------------------------------------------------------------------------------------------

# [F] - Fragments
class F(Feature):
    symbol = 'F'
    name = 'Fragments'
    feature_enum = FeatureEnum.F

    @classmethod
    def is_standardized(cls, mol: Mol) -> bool:
        n_frags = rdk.GetMolFrags(mol.mol)
        return len(n_frags) <= 1

    @classmethod
    def standardise_rdk(cls, rdk_mol: rdk.Mol) -> rdk.Mol:
        mol_ = rdk_std.FragmentParent(rdk_mol, skipStandardize = True)
        return mol_


# [I] - Isotope
class I(Feature):
    symbol = 'I'
    name = 'Isotope'
    feature_enum = FeatureEnum.I

    @classmethod
    def is_standardized(cls, mol: Mol) -> bool:
        a: rdk.Atom
        for a in mol.mol.GetAtoms():
            if a.GetIsotope():
                return False
        return True

    @classmethod
    def standardise_rdk(self, rdk_mol: rdk.Mol) -> rdk.Mol:
        mol_ = copy(rdk_mol)
        a: rdk.Atom
        for a in mol_.GetAtoms():
            if a.GetIsotope():
                a.SetIsotope(0)
        return mol_


# [C] - Charge
class C(Feature):
    symbol = 'C'
    name = 'Charge'
    feature_enum = FeatureEnum.C

    # TODO use other uncharging approach?
    uncharger = rdk_std.Uncharger()

    @classmethod
    def is_standardized(cls, mol: Mol) -> bool:
        mol_ = cls.uncharger.uncharge(mol.mol)
        return mol_identical(mol.mol, mol_)

    @classmethod
    def standardise_rdk(cls, rdk_mol: rdk.Mol) -> rdk.Mol:
        mol_ = cls.uncharger.uncharge(rdk_mol)
        return mol_


# [T] - Tautomer
class T(Feature):
    symbol = 'T'
    name = 'Tautomer'
    feature_enum = FeatureEnum.T

    tautomer_enumerator = rdk_std.TautomerEnumerator()

    @classmethod
    def is_standardized(cls, mol: Mol) -> bool:
        """ Checks if the generated canonical tautomer is different. """
        try:
            mol_canon = cls.tautomer_enumerator.Canonicalize(mol.mol)
            return mol_identical(mol.mol, mol_canon)
        except rdk.rdchem.AtomKekulizeException:
            return False

    @classmethod
    def standardise_rdk(cls, rdk_mol: rdk.Mol) -> rdk.Mol:
        mol_ = cls.tautomer_enumerator.Canonicalize(rdk_mol)
        return mol_


# [S] - Stereo
class S(Feature):
    symbol = 'S'
    name = 'Stereo'
    feature_enum = FeatureEnum.S

    @classmethod
    def is_standardized(cls, mol: Mol) -> bool:
        # TODO redo this check
        return not ('/' in mol.smiles or '\\' in mol.smiles or '@' in mol.smiles)

    @classmethod
    def standardise_rdk(cls, rdk_mol: rdk.Mol) -> rdk.Mol:
        mol_ = copy(rdk_mol)
        rdk.RemoveStereochemistry(mol_)
        return mol_


feat_enum_map: dict[FeatureEnum, Type[Feature]] = {
    FeatureEnum.F: F,
    FeatureEnum.I: I,
    FeatureEnum.C: C,
    FeatureEnum.T: T,
    FeatureEnum.S: S,
}

# List of features
feats: list[Type[Feature]] = [F, I, C, T, S]


# Utils
# --------------------------------------------------------------------------------------------------------------
def mol_identical(a: rdk.Mol, b: rdk.Mol) -> bool:
    return a.HasSubstructMatch(b) and b.HasSubstructMatch(a)






from typing import Optional

from chemrecon.schema.db_object import Column, Relation, InverseRelation
from chemrecon.schema.entry_types.aam_repr import AAMRepr
from chemrecon.schema.entry_types.molstructure_repr import MolStructureRepr


class AAMReprInvolvesMolStructureRepr(Relation[AAMRepr, MolStructureRepr]):
    """ Relates the molecular structure representation involved in an atom-to-atom-mapped reaction.
        These molecular structures are implicitly a part of the AAM structure.
    """
    # Attributes
    index: int  #: Index in the mapping string
    stoich: int  #: The stoichiometric coefficient of the MolStructureRepr

    # Database
    entrytype_name = 'AAMRepr Involves MolStructureRepr'
    _table_name = 'AAMReprInvolvesMolStructureRepr'
    symmetric = False
    source_entrytype = AAMRepr
    target_entrytype = MolStructureRepr
    _attribute_columns = [
        Column('index', int),
        Column('stoich', int)
    ]
    _index = [0]

    def __init__(
        self,
        index: int,
        stoich: int,
        recon_id_1: Optional[int] = None,
        recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.index = index
        self.stoich = stoich

    def _vis_str(self) -> str:
        return f'has structure (side: {self.stoich})'

class MolStructureReprParticipatesInAAMRepr(InverseRelation[MolStructureRepr, AAMRepr]):
    """ Relates the molecular structure representation involved in an atom-to-atom-mapped reaction.
        These molecular structures are implicitly a part of the AAM structure.
    """
    # Attributes
    index: int  #: Index in the mapping string
    stoich: int  #: The stoichiometric coefficient of the MolStructureRepr

    # Inverse
    inverse_main_relation = AAMReprInvolvesMolStructureRepr

    # Database
    entrytype_name = 'MolStructureRepr Participates in AAMRepr'
    _table_name = 'MolStructureReprParticipatesInAAMRepr'
    symmetric = False
    source_entrytype = MolStructureRepr
    target_entrytype = AAMRepr
    _attribute_columns = [
        Column('index', int),
        Column('stoich', int)
    ]
    _index = [0]

    def __init__(
        self,
        index: int,
        stoich: int,
        recon_id_1: Optional[int] = None,
        recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.index = index
        self.stoich = stoich

    def _vis_str(self) -> str:
        return f'in aam ({self.stoich})'

AAMReprInvolvesMolStructureRepr.has_inverse = MolStructureReprParticipatesInAAMRepr



""" Drawing utilites for each type of entry.
"""
from __future__ import annotations

from typing import Type, Optional

from PIL import Image
import rustworkx.visualization as rx_vis

import chemrecon.entrygraph.entrygraph
from chemrecon.schema import Entry, Relation

attr_none: str = "''"
linewidth: int = 30


# noinspection PyProtectedMember
class EntryGraphDrawer():
    entrygraph_type: Type[chemrecon.entrygraph.entrygraph.EntryGraph]

    # Draw settings
    draw_reconids: bool

    def __init__(
        self,
        # Settings
        draw_reconids: bool = False
    ):
        """ The initializer for the Drawer sets the settings
        """
        self.draw_reconids = draw_reconids

    def draw(
        self,
        entrygraph: chemrecon.entrygraph.entrygraph.EntryGraph,
        filename: str = None,
        filetype: str = 'png',
        scores: Optional[dict[Entry, float]] = None
    ) -> Image.Image:
        """ Draw the entry graph.

            For possible file formats, see the the corresponding
            [RustWorkX documentation](https://www.rustworkx.org/apiref/rustworkx.visualization.graphviz_draw.html).
        """

        g_ = entrygraph.g.copy()

        # Empty score dict if not given
        score_dict: dict[Entry, float] = scores if scores is not None else dict()

        # Remove duplicate edges (for symmetric relations)
        for e_index in g_.edge_indices():
            data = g_.get_edge_data_by_index(e_index)
            if data.relation.symmetric:
                x, y = g_.get_edge_endpoints_by_index(e_index)
                if x > y:
                    g_.remove_edge_from_index(e_index)

        # Remove self-edges
        for e_index in g_.edge_indices():
            x, y = g_.get_edge_endpoints_by_index(e_index)
            if x == y:
                g_.remove_edge_from_index(e_index)

        return rx_vis.graphviz_draw(
            g_,
            node_attr_fn = lambda v: self.node_attr(v.entry, score_dict.get(v.entry, None)),
            edge_attr_fn = lambda e: self.edge_attr(e.relation),
            graph_attr = None,  # TODO
            filename = filename,
            image_type = filetype,
        )

    def node_attr(self, entry: Entry, score: Optional[float]) -> dict[str, str]:
        assert entry.recon_id is not None

        # Label
        if self.draw_reconids:
            recon_id_part = f'\n{entry.recon_id}'
        else:
            recon_id_part = ''

        # Break lines of label
        entrylabel = entry._vis_str()
        label_lines = entrylabel.splitlines()
        split_label: list[str] = list()
        for l in label_lines:
            while l != '':
                split_label.append(l[:linewidth])
                l = l[linewidth:]
        entry_label_final = ('\n').join(split_label)

        # Determine label
        label: str = f'{entry.get_table_name()}\n{entry_label_final}{recon_id_part}'
        if score is not None:
            label += f'\n{score:.3f}'

        # Other drawing
        d = {
            'style': 'filled',
            'shape': 'box',
            'label': label
        }
        d.update(entry._vis_attrs())
        return d

    def edge_attr(self, rel: Relation) -> dict[str, str]:
        label = rel._vis_str()
        d = {
            'label': label,
            'dir': 'none' if rel.symmetric else 'forward',
        }

        # Update with rel_type-specific attributes and return
        d.update(rel._vis_attrs())
        return d

# Default drawer - this is used by the EntryGraph .draw() and .show() methods.
defaultdrawer = EntryGraphDrawer()



import networkx as nx

def max_weight_matching[T1, T2](
    edges: dict[tuple[T1, T2], float],
    min_weight: bool = False
) -> dict[T1, T2]:
    """ Takes as edges a dictionary T1 -> [T2], with a float value giving the weight for each match in the
        second set.
        Returns the maximum weight bipartite matching as a list of tuples (T1, T2).
        If min_weight is true, uses negative weight values.
    """
    # Construct a weighted bipartite graph
    g = nx.Graph()
    for (a, b), w in edges.items():
        g.add_node(a)
        g.add_node(b)
        g.add_edge(a, b, weight = ((2 - w) if min_weight else w))

    # Find the maximal matching
    matching = nx.max_weight_matching(g, maxcardinality = True)
    keys: set[T1] = {k for k, _ in edges.keys()}
    out: dict[T1, T2] = dict()
    for a, b in matching:
        if a in keys:
            out[a] = b
        elif b in keys:
            out[b] = a
        else:
            raise RuntimeError

    return out



from typing import Optional

from chemrecon.schema.db_object import Column, Relation, InverseRelation
from chemrecon.schema.entry_types.aam import AAM
from chemrecon.schema.entry_types.molstructure import MolStructure


class AAMInvolvesMolStructure(Relation[AAM, MolStructure]):
    """ Relates the molecular structures involved in an atom-to-atom-mapped reaction.
        These molecular structures are implicitly a part of the AAM structure.
    """
    # Attributes
    side: int  #: The side of the reaction which contains this structure (-1 or 1)
    structure_atom_index_in_aam: list[int]  #: Atom mapping numbers of the atoms in the structure w.r.t. the AAM.

    # Database
    entrytype_name = 'AAM Involves MolStructure'
    _table_name = 'AAMInvolvesMolStructure'
    symmetric = False
    source_entrytype = AAM
    target_entrytype = MolStructure
    _attribute_columns = [
        Column('structure_atom_index_in_aam', list[int])
    ]
    _index = [0]

    def __init__(
        self,
        side: int,
        structure_atom_index_in_aam: list[int],
        recon_id_1: Optional[int] = None,
        recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.side = side
        self.structure_atom_index_in_aam = structure_atom_index_in_aam

    def _vis_str(self) -> str:
        return f'has structure ({self.side})'

class MolStructureParticipatesInAAM(InverseRelation[MolStructure, AAM]):
    """ Relates the molecular structures involved in an atom-to-atom-mapped reaction.
        These molecular structures are implicitly a part of the AAM structure.
    """
    # Attributes
    side: int  #: The side of the reaction which contains this structure (-1 or 1)
    structure_atom_index_in_aam: list[int]  #: Atom mapping numbers of the atoms in the structure w.r.t. the AAM.

    # Inverse
    inverse_main_relation = AAMInvolvesMolStructure

    # Database
    entrytype_name = 'MolStructure Participats in AAM'
    _table_name = 'MolStructureParticipatesInAAM'
    symmetric = False
    source_entrytype = MolStructure
    target_entrytype = AAM
    _attribute_columns = [
        Column('structure_atom_index_in_aam', list[int])
    ]
    _index = [0]

    def __init__(
        self,
        side: int,
        structure_atom_index_in_aam: list[int],
        recon_id_1: Optional[int] = None,
        recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.side = side
        self.structure_atom_index_in_aam = structure_atom_index_in_aam

    def _vis_str(self) -> str:
        return f'in reaction ({self.side})'

AAMInvolvesMolStructure.has_inverse = MolStructureParticipatesInAAM



""" Re-exports all concrete schema objects
"""
from __future__ import annotations

from chemrecon.schema.db_object import (Entry, SourceEntry, Relation, DatabaseObject, Column,
                                        ProceduralRelation, ProceduralGeneratorError, ComposedRelation)

from chemrecon.schema.enums import *

# Re-export entry types
from chemrecon.schema.entry_types.aam import AAM
from chemrecon.schema.entry_types.aam_repr import AAMRepr
from chemrecon.schema.entry_types.compound import Compound
from chemrecon.schema.entry_types.enzyme import Enzyme
from chemrecon.schema.entry_types.reaction import Reaction
from chemrecon.schema.entry_types.molstructure_repr import MolStructureRepr
from chemrecon.schema.entry_types.molstructure import MolStructure

# Re-export relation types (with corresponding inverses, defined in the same file as the main)
from chemrecon.schema.relation_types_source.aam_involves_molstructure_relation import *
from chemrecon.schema.relation_types_source.aam_repr_involves_molstructure_repr_relation import *
from chemrecon.schema.relation_types_procedural.aam_convert_relation import *
from chemrecon.schema.relation_types_composed.compound_has_molstructure_relation import *
from chemrecon.schema.relation_types_source.compound_has_structure_representation_relation import *
from chemrecon.schema.relation_types_source.compound_reference_relation import *
from chemrecon.schema.relation_types_composed.reaction_has_aam_relation import *
from chemrecon.schema.relation_types_source.reaction_has_aam_representation_relation import *
from chemrecon.schema.relation_types_source.reaction_has_enzyme_relation import *
from chemrecon.schema.relation_types_source.reaction_involves_compound_relation import *
from chemrecon.schema.relation_types_source.reaction_reference_relation import *
from chemrecon.schema.relation_types_procedural.molstructure_convert_relation import *
from chemrecon.schema.relation_types_source.molstructure_standardisation_relation import *

# Ontology
from chemrecon.schema.relation_types_source.ontology.compound_ontology import *
from chemrecon.schema.relation_types_source.ontology.reaction_ontology import *
from chemrecon.schema.relation_types_source.ontology.enzyme_ontology import *

# Procedural relation types, import only the relation, not the protocol
from chemrecon.schema.relation_types_procedural.compound_select_structure_proceduralrelation import (
    CompoundSelectStructure
)
from chemrecon.schema.relation_types_procedural.reaction_select_aam_proceduralrelation import (
    ReactionSelectAAM
)

# Export lists
entrytypes: list[type[Entry]] = [
    Compound, MolStructureRepr, MolStructure, Reaction, Enzyme,
    AAMRepr, AAM
]
relationtypes: list[type[Relation]] = [
    # Main source relations
    CompoundReference,
    CompoundHasStructureRepresentation, CompoundHasMolStructure,
    MolStructureConvert, MolStructureStandardization,
    ReactionReference,
    ReactionInvolvesCompound, CompoundParticipatesInReaction,
    ReactionHasEnzyme,
    ReactionHasAAMRepr,
    ReactionHasAAM,
    AAMReprInvolvesMolStructureRepr,
    AAMConvert,
    AAMInvolvesMolStructure,

    # Ontology, Compound
    CompoundIsA, CompoundHasInstance,
    CompoundHasNewID, CompoundHasOldID,
    CompoundHasPart, CompoundIsPartOf,
    CompoundHasConjugateAcid, CompoundHasConjugateBase,
    CompoundHasTautomer,
    CompoundHasStereoIsomer,
    CompoundHasIsotopologue,

    # Ontology, Reaction
    ReactionIsA, ReactionHasInstance,
    ReactionHasNewID, ReactionHasOldID,

    # Ontology, Enzyme
    EnzymeIsA, EnzymeHasInstance,
    EnzymeHasNewID, EnzymeHasOldID,

    # EntryGraph Procedural
    CompoundSelectStructure,
    ReactionSelectAAM,
]



""" Implements a wrapper for the RDKit reaction type."""
from __future__ import annotations

from enum import Enum
from typing import Optional

import rdkit.Chem.rdChemReactions as rdk_r
from rdkit.Chem.rdChemReactions import ChemicalReaction
from rdkit.Chem import Draw as rdk_draw

from chemrecon.chem.mol import MolTemplate, MolInstance
from chemrecon.chem.sumformula import SumFormula
from chemrecon.schema.entry_types.aam import AAM


class Side(Enum):
    L = -1
    R = 1

type stoich = int

class ChemReaction:
    reaction: ChemicalReaction
    _reaction_smiles: Optional[str]

    # Templates represent the structure of each compound. (There can be duplicate templates)
    lhs_templates: dict[MolTemplate, stoich]
    rhs_templates: dict[MolTemplate, stoich]

    # Instances represent specific structures with maps as they participate in the reaction.
    lhs_instances: list[MolInstance]
    rhs_instances: list[MolInstance]

    # Mapping between each instance and its corresponding template
    instance_template_dict = dict[MolInstance, MolTemplate]

    # Properties of the given instances, as read through the input files if possible
    instance_properties: dict[MolInstance, dict[str, str]]
    template_ids: dict[MolTemplate, str]
    template_names: dict[MolTemplate, str]

    # Atom-to-atom map of this reaction
    map: dict[tuple[MolInstance, int], tuple[MolInstance, int]]
    lhs_index: dict[int, tuple[MolInstance, int]]  # The global map number to (molecule, index) tuple
    rhs_index: dict[int, tuple[MolInstance, int]]  # The global map number to (molecule, index) tuple

    def __init__(self, rdk_reaction: ChemicalReaction):
        self.reaction = rdk_reaction
        self._reaction_smiles = None

        self.lhs_instances = list()
        self.rhs_instances = list()
        self.lhs_templates = dict()
        self.rhs_templates = dict()

        self.instance_template_dict = dict()
        self.map = dict()

        self.instance_properties = dict()
        self.template_ids = dict()
        self.template_names = dict()

        # Set reactant and product MolInstances and map
        for rdk_mol in self.reaction.GetReactants():
            self.lhs_instances.append(
                MolInstance(rdk_mol, provenance = 'implicit')
            )

        for rdk_mol in self.reaction.GetProducts():
            self.rhs_instances.append(
                MolInstance(rdk_mol, provenance = 'implicit')
            )

        # Populate map
        # The global map number to (molecule, index) tuple
        # Global map number -> molecule, index (referring to the order of atoms in the SMILES string)
        self.lhs_index = dict()
        self.rhs_index = dict()
        for lhs_instance in self.lhs_instances:
            for local_index, global_index in enumerate(lhs_instance.get_atom_map_in_native_order()):
                self.lhs_index[global_index] = (lhs_instance, local_index)
        for rhs_instance in self.rhs_instances:
            for local_index, global_index in enumerate(rhs_instance.get_atom_map_in_native_order()):
                self.rhs_index[global_index] = (rhs_instance, local_index)

        self.map = dict()
        for i, (lhs_mol, lhs_index) in self.lhs_index.items():
            try:
                self.map[(lhs_mol, lhs_index)] = self.rhs_index[i]
            except KeyError:
                # Missing map
                # TODO do something if this is not a hydrogen?
                pass

        # self.map = {
        #     (lhs_mol, lhs_index): self.rhs_index[i]
        #     for i, (lhs_mol, lhs_index) in self.lhs_index.items()
        # }

        # Compute templates and add
        for side, instance_list, template_list in [
            (-1, self.lhs_instances, self.lhs_templates),
            (1, self.rhs_instances, self.rhs_templates)
        ]:
            mol_instance: MolInstance
            for mol_instance in instance_list:
                template = mol_instance.to_mol_template()

                # Add to instance-template dict
                self.instance_template_dict[mol_instance] = template

                # Add to template lists
                template_list[template] = template_list.get(template, 0) + side

    # Getters
    # ------------------------------------------------------------------------------------------------------------------
    def get_lhs_templates(self) -> list[MolTemplate]:
        return list(self.lhs_templates.keys())

    def get_rhs_templates(self) -> list[MolTemplate]:
        return list(self.rhs_templates.keys())

    # Balance
    # ------------------------------------------------------------------------------------------------------------------
    def get_balance_difference(self) -> SumFormula:
        """ Get the difference between the LHS and RHS as a (possible negative) MolFormula.
            Positive counts indicate a surplus on the LHS, negative counts indicate a surplus on the RHS.
        """
        raise NotImplementedError()

        # lhs_sum = sum(m.get_molformula() for m in self.lhs_instances)
        # rhs_sum = sum(m.get_molformula() for m in self.rhs_instances)
        # return lhs_sum - rhs_sum

    def is_balanced(self) -> bool:
        """ Returns true if balanced in both atomic composition and charge.
            If not balanced, use .get_balance_differenec() to inspect the difference between LHS and RHS.
        """
        return self.get_balance_difference().is_zero()

    # Representations
    # ------------------------------------------------------------------------------------------------------------------
    def to_reaction_smiles(self) -> str:
        if self._reaction_smiles is None:
            # self._reaction_smiles = rdk_r.ReactionToSmarts(self.reaction)
            self._reaction_smiles = rdk_r.ReactionToSmiles(self.reaction)

        return self._reaction_smiles

    # Serialise
    def serialize(self) -> dict:
        return {
            'reaction_smiles': self.to_reaction_smiles(),
            'lhs': [
                mol.serialize() for mol in self.lhs_instances
            ],
            'rhs': [
                mol.serialize() for mol in self.rhs_instances
            ],
            'balanced': 'TODO',  # TODO
            'balance_difference': 'TODO',  # TODO compute difference in sum formulae for LHS and RHS
        }

    # Misc
    # ------------------------------------------------------------------------------------------------------------------
    def __hash__(self):
        return self.to_reaction_smiles().__hash__()

    def sanity_check(self):
        """ Raises an exception if the AAM maps atoms of different elements.
        """
        # Check all instances have smiles
        for inst in [*self.lhs_instances, *self.rhs_instances]:
            if inst.smiles is None:
                raise AssertionError('Instance invalid.')

        # Check mapping
        for global_index, (molinst_l, local_index_l) in self.lhs_index.items():
            try:
                molinst_r, local_index_r = self.rhs_index[global_index]
            except KeyError:
                # Atom 'disappears'
                continue

            l_atom = molinst_l.mol.GetAtomWithIdx(local_index_l)
            r_atom = molinst_r.mol.GetAtomWithIdx(local_index_r)
            if l_atom.GetAtomicNum() != r_atom.GetAtomicNum():
                raise AssertionError(f'Element mismatch: L: {local_index_l}, R: {local_index_r}')




    # Visualisation
    # ------------------------------------------------------------------------------------------------------------------
    def show(self):
        img = rdk_draw.ReactionToImage(rxn = self.reaction, subImgSize = (800, 800))
        img.show()
        pass

# Creators
# ----------------------------------------------------------------------------------------------------------------------
def chem_reaction_from_aam_entry(
    entry: AAM
) -> ChemReaction:
    """ Given an AAM entry, load into a ChemReaction object.
    """
    return chem_reaction_from_reactionsmiles(
        entry.reaction_smiles
    )

def chem_reaction_from_reactionsmiles(
    reactionsmiles: str
) -> ChemReaction:
    rdk_reaction = rdk_r.ReactionFromSmarts(reactionsmiles)
    return ChemReaction(rdk_reaction)

def chem_reaction_from_rxn(
    rxn: str,
    provenance: Optional[str] = None,
    safe: bool = False
) -> ChemReaction:
    rdk_reaction = rdk_r.ReactionFromRxnBlock(rxn, sanitize = not safe, removeHs = not safe)
    return ChemReaction(rdk_reaction)



from typing import Optional

from chemrecon.schema import Column, Entry, IdTypeStructureRepresentationEnum
from chemrecon.schema.db_object import col_recon_id, col_source_id_hashed
class MolStructureRepr(Entry):
    """ Represents a representation of a structure, such as an S_SMILES string, S_MOLFILE or similar. """
    # Attributes
    source_id: str  #: Identifier string. (**index**)
    id_type: IdTypeStructureRepresentationEnum  #: The type of the identifier string (e.g. SMILES, InChI). (**index**)
    #: If marked as implicit, this molecular structure only exists implicitly in an atom-to-atom map, and is not
    #: directly referenced by a known compound.
    implicit: bool

    # Database
    entrytype_name = 'MolStructureRepr'
    _table_name = 'MolStructureRepr'
    _columns = [
        col_recon_id,
        col_source_id_hashed,
        Column('id_type', IdTypeStructureRepresentationEnum),
        Column('implicit', bool)
    ]
    _index = [1, 2]

    # Visualisation
    _draw_colour = '7AFF78'

    def __init__(
        self,
        source_id: str,
        implicit: bool,
        id_type: IdTypeStructureRepresentationEnum,
        recon_id: Optional[int] = None,
    ):
        super().__init__(recon_id)
        self.source_id = source_id
        self.id_type = id_type
        self.implicit = implicit

    def _vis_str(self) -> str:
        return f'{self.id_type.name}: {self.source_id}'



""" Defines metadata and the most general exports in the chemrecon.* namespace.
"""
# Metadata
__version__ = '0.1.1'  # Library version
__db_version__: list[str] = ['0.1.1']  # Compatible database versions


# Imports
import psycopg


# Connection and handler
import chemrecon.connection
from chemrecon.connection import (
    connect, connect_public, connect_local_docker, connect_local_docker_dev, disconnect,
    get_query_handler
)
from chemrecon.database.params import Params


# Project initialization
# Try to set the default database connection
try:
    disconnect()
    # connect_public()
except psycopg.OperationalError as e:
    print(f'Cannot connect to default database: {e}')

# Import from schema
from chemrecon.schema import *

from chemrecon.core.query_handler import QueryHandler
from chemrecon.core.populate_query_handler import PopulateQueryHandler

# Export identifier types
from chemrecon.core.id_types import *

# Entry creators
# from chemrecon.query.create_entry import (
#     entry,
#     compound_entry, reaction_entry, enzyme_entry,
#     aam_representation_entry, structure_representation_entry,
#     structure_entry, aam_entry,
#     enzyme_from_ec_number,
#     entry_from_identifiers_org
# )

from chemrecon.query.find_entry import (
    find_entry,
    find_compound_entry, find_reaction_entry, find_enzyme_entry,
    find_structure_representation_entry, find_aam_representation_entry,
    find_structure_entry, find_aam_entry
)

# Relation getters
from chemrecon.query.get_relations import *

# EntryGraphs
from chemrecon.entrygraph.entrygraph import EntryGraph, Vertex, Edge
from chemrecon.schema.direction import Direction
from chemrecon.entrygraph.filter import (
    EntryFilter, EntryFilterProcedure,
    RelationFilter, RelationFilterProcedure
)
from chemrecon.entrygraph.explorationprotocol import ExplorationProtocol
from chemrecon.entrygraph.explore import explore
from chemrecon.entrygraph.scoring import Scorer

# Pre-defined entrygraph types
from chemrecon.query.default_protocols import *

# Chemistry - Molecules
from chemrecon.chem.mol import Mol



from typing import Optional

from chemrecon.schema import Entry, IdTypeAAMEnum
from chemrecon.schema.db_object import Column, col_recon_id, col_source_id_hashed


class AAMRepr(Entry):
    """ Represents an enzyme entry. """
    # Attributes
    source_id: str  #: Identifier string. (**index**)
    id_type: IdTypeAAMEnum  #: The type of the identifier string (e.g. Reaction SMILES). (**index**)

    # Database
    entrytype_name = 'AAMRepr'
    _table_name = 'AAMRepr'
    _columns = [
        col_recon_id,
        col_source_id_hashed,
        Column('id_type', IdTypeAAMEnum),
    ]
    _index = [1, 2]

    # Visualisation
    _draw_colour = '7AFF78'

    def __init__(
        self,
        source_id: str,
        id_type: IdTypeAAMEnum,
        recon_id: Optional[int] = None,
    ):
        super().__init__(recon_id)
        self.source_id = source_id
        self.id_type = id_type

    def _vis_str(self) -> str:
        return f'{self.id_type.name}: {self.source_id}'


from typing import Optional

import psycopg as pg
import psycopg.sql as sql

from chemrecon.schema.db_object import Entry, Relation
import chemrecon.core.query_handler

class PopulateQueryHandler(chemrecon.core.query_handler.QueryHandler):
    """ Entries and relations are added to a PopulationQueryHandler, which batches operations and
        sends them to the database.
        When added, the handler will assign a correct ReconID to the object based on the current state of the
        corresponding sequence.
        Should be used as a context.
    """

    # Adding entries to the handler
    # ----------------------------------------------------------------------------------------------------------
    # For adding entries when the assigned reconid is needed. This cannot be batched.
    def add_entry[T: Entry](
        self,
        entry: T,
        returning: bool = True,
        authoritative: bool = False
    ) -> Optional[int]:
        """ Add an entry to the DB, returning its reconid immediately.
            If authoritative, allow overriding fields added previously.
            If not authoritative, do nothing on conflict.
        """
        if returning:
            return self.add_entries([entry], returning = returning, authoritative = authoritative)[0]
        else:
            self.add_entries([entry], returning = returning, authoritative = authoritative)
            return None

    def add_entries[T: Entry](
        self,
        entries: list[T],
        returning: bool = True,
        authoritative: bool = False
    ) -> Optional[list[int]]:
        """ TODO
        """
        entrytype = type(entries[0])
        cols = entrytype.get_columns(include_recon_id = False)
        index_cols = entrytype.get_index_columns()
        table_identifier = sql.Identifier(entrytype.get_table_name())

        # TODO 'returning' clause can be removed if not returning?
        q = sql.SQL("""
            INSERT INTO {entrytypetable} ({columns})
            VALUES ({placeholders})
                ON CONFLICT ({index_columns}) 
                DO UPDATE SET ({columns}) = ROW({coalesced_values})
             RETURNING {reconid_col};
        """).format(
            entrytypetable = table_identifier,
            reconid_col = sql.Identifier('recon_id'),
            columns = sql.SQL(',').join(
                sql.Identifier(c.name) for c in cols
            ),
            index_name = sql.Identifier(f'{entrytype.get_table_name()}_index'),
            index_columns = sql.SQL(',').join(  # TODO better way with named constraint?
                sql.Identifier(c.name) if not c.index_hash
                else sql.SQL('md5({colname})').format(colname = sql.Identifier(c.name))
                for c in index_cols
            ),
            placeholders = sql.SQL(', ').join([sql.SQL('%s') for _ in cols]),
            coalesced_values = sql.SQL(',').join(
                # Coalesce values, preferring existing entry
                sql.SQL('COALESCE(excluded.{col}, {table}.{col})').format(
                    col = sql.Identifier(c.name),
                    table = table_identifier
                )
                for c in cols
            )
        )

        # Execute
        try:
            self.c.executemany(
                q,
                params_seq = [
                    list(e.get_columns_with_values(include_recon_id = False).values())
                    for e in entries
                ],
                returning = returning
            )

            # Return as dict
            if returning:
                return [i[0] for i in self._extract_results_from_cursor()]
            else:
                return None
        except pg.errors.UniqueViolation as e:
            raise KeyError(f'Unique violation: {e}\n\t{e}')
        except pg.errors.SyntaxError as e:
            print(f'ERROR:\n{q.as_string()}\n\t{e}')
            raise RuntimeError

    def add_relation[T1: Entry, T2: Entry](self, relation: Relation[T1, T2]):
        """ When adding a relation, it is expected that the 'entry_1' and 'entry_2' fields are filled such that
            the appropriate key can be looked up.
        """
        self.add_relations([relation])

    def add_relations[T1: Entry, T2: Entry](self, relations: list[Relation[T1, T2]]):
        # TODO dispatch to add by reconid or by entries, depending
        # TODO change to the plural version, specialised for adding only one
        raise NotImplementedError()

    def add_relation_by_recon_ids[T1: Entry, T2: Entry](
        self,
        recon_id_1: int,
        recon_id_2: int,
        relation: Relation[T1, T2]
    ):
        """ Add a relation, using the recon_id fields of the given relation.
        """
        self.add_relations_by_recon_ids([(recon_id_1, recon_id_2, relation)])

    def add_relations_by_recon_ids[T1: Entry, T2: Entry](
        self,
        relations: list[tuple[int, int, Relation[T1, T2]]]
    ):
        """ Add relations by recon_ids. """
        reltype = type(relations[0][2])

        # Add recon_ids to the relation, sort recon_ids of needed
        if reltype.symmetric:
            for r in relations:
                r[2].recon_id_1, r[2].recon_id_2 = sorted((r[0], r[1]))
        else:
            for r in relations:
                r[2].recon_id_1, r[2].recon_id_2 = r[0], r[1]

        cols = reltype.get_columns()
        q = sql.SQL("""
                    INSERT INTO {rel_table_name} ({cols})
                    VALUES ({placeholders})
                        ON CONFLICT DO NOTHING;
                    """).format(
            rel_table_name = sql.Identifier(reltype.get_table_name()),
            cols = sql.SQL(', ').join(sql.Identifier(c.name) for c in cols),
            placeholders = sql.SQL(', ').join([sql.SQL('%s') for _ in cols])
        )
        self.c.executemany(
            q,
            params_seq = [
                list(r[2].get_columns_with_values().values())
                for r in relations
            ]
        )

    def add_relation_by_entries[T1: Entry, T2: Entry](
        self,
        entry_1: T1,
        entry_2: T2,
        relation: Relation[T1, T2]
    ):
        """ Add a relation, using the entry fields of the given relation.
        """
        self.add_relations_by_entries([(entry_1, entry_2, relation)])

    def add_relations_by_entries[T1: Entry, T2: Entry](
        self,
        relations: list[tuple[T1, T2, Relation[T1, T2]]]
    ):
        """ Add relations using the entry fields of each given relation.
        """
        reltype = type(relations[0][2])
        all_rel_cols = reltype.get_columns()
        attrcols = reltype.get_attribute_columns()
        e1_table, e2_table = reltype.get_entry_table_names()
        e1_index_cols = reltype.source_entrytype.get_index_columns()
        e2_index_cols = reltype.target_entrytype.get_index_columns()

        if reltype.symmetric:
            q_ = sql.SQL("""
                 WITH r_id1 AS (
                     SELECT recon_id
                     FROM {entry_1_table}
                     WHERE ({entry_1_index_cols}) = ({entry_1_placeholders})
                 ), r_id2 AS (
                     SELECT recon_id
                     FROM {entry_2_table}
                     WHERE ({entry_2_index_cols}) = ({entry_2_placeholders})
                 )
                 INSERT INTO {rel_table_name} ({all_columns})
                 VALUES (
                     LEAST((SELECT recon_id FROM r_id1), (SELECT recon_id FROM r_id2)),
                     GREATEST((SELECT recon_id FROM r_id1), (SELECT recon_id FROM r_id2))
                     {attr_placeholders}
                     )
                 ON CONFLICT DO NOTHING;
            """)
        else:
            q_ = sql.SQL("""
                 WITH r_id1 AS (
                     SELECT recon_id
                     FROM {entry_1_table}
                     WHERE ({entry_1_index_cols}) = ({entry_1_placeholders})
                 ), r_id2 AS (
                     SELECT recon_id
                     FROM {entry_2_table}
                     WHERE ({entry_2_index_cols}) = ({entry_2_placeholders})
                 )
                 INSERT INTO {rel_table_name} ({all_columns})
                 VALUES ((SELECT recon_id FROM r_id1), (SELECT recon_id FROM r_id2) {attr_placeholders})
                 ON CONFLICT DO NOTHING;
            """)

        # Prepare
        q = q_.format(
            # Entry 1
            entry_1_table = sql.Identifier(e1_table),
            entry_1_index_cols = sql.SQL(', ').join(
                sql.Identifier(c.name) for c in e1_index_cols
            ),
            entry_1_placeholders = sql.SQL(', ').join(sql.SQL('%s') for _ in e1_index_cols),
            # Entry2
            entry_2_table = sql.Identifier(e2_table),
            entry_2_index_cols = sql.SQL(', ').join(
                sql.Identifier(c.name) for c in e2_index_cols
            ),
            entry_2_placeholders = sql.SQL(', ').join(sql.SQL('%s') for _ in e2_index_cols),
            # Attributes
            rel_table_name = sql.Identifier(reltype.get_table_name()),
            all_columns = sql.SQL(', ').join(
                sql.Identifier(c.name) for c in all_rel_cols
            ),
            attr_placeholders = (
                    sql.SQL(', ') + sql.SQL(', ').join(sql.SQL('%s') for _ in attrcols)
            ) if len(attrcols) > 0 else sql.SQL('')
        )

        # Execute
        self.c.executemany(
            q,
            params_seq = [
                [
                    *e1.get_index_columns_with_values().values(),
                    *e2.get_index_columns_with_values().values(),
                    *r.get_columns_with_values(include_recon_ids = False).values()
                ]
                for e1, e2, r in relations
            ],
            returning = False
        )

    # Connected adders
    def add_entry_with_relations[T1: Entry, T2: Entry](
        self, entry: T1, relations: list[tuple[Relation[T1, T2], T2]]
    ) -> int:
        """ An entry and a number of relations to possibly new entries with the first entry as the source.
            Relations do not need to have their recon_id_1, recon_id_2, or entries filled.
            Returns id of source entry
        """
        # Add the T1 entry
        e_recon_id = self.add_entry(entry, returning = True)

        if len(relations) == 0:
            # Short circuit if no relations given
            return e_recon_id

        rels, t2_entries = zip(*relations)
        rel_type = type(rels[0])
        t2_type = type(t2_entries[0])

        # Add relations and secondary entries
        self.add_relations_to_entry_with_reconid(
            recon_id = e_recon_id,
            entry_table = type(entry),
            relations = relations
        )

        # Return id of added/found source entry
        return e_recon_id


    def add_relations_to_entry_with_reconid[T1: Entry, T2: Entry](
        self,
        recon_id: int,
        entry_table: type[Entry],
        relations: list[tuple[Relation[T1, T2], T2]] | list[tuple[Relation[T2, T1], T2]]
    ) -> list[int]:
        """ Add a number of relations from the given reconid.
            Returns a list of assigned Recon IDs in the same order as the input.
        """
        if len(relations) == 0:
            return []

        # Gather types
        rels, t2_entries = zip(*relations)
        rel_type = type(rels[0])
        t2_type = type(t2_entries[0])

        # Add all T2 entries
        t2_ids: list[int] = self.add_entries(t2_entries, returning = True, authoritative = False)
        if len(t2_ids) != len(relations):
            raise ValueError

        # Dispatch, setting recon_ids of added relations depending on type
        if rel_type.symmetric:
            # T2 = T1, Relation[T1, T1]
            for t2_id, rel in zip(t2_ids, rels):
                rel: Relation[T1, T1]
                rel.recon_id_1 = recon_id
                rel.recon_id_2 = t2_id
                rel.recon_id_1, rel.recon_id_2 = sorted((rel.recon_id_1, rel.recon_id_2))  # Keep invariant

        elif rel_type.source_entrytype is entry_table:
            # Relation[T1, T2]
            for t2_id, rel in zip(t2_ids, rels):
                rel: Relation[T1, T2]
                rel.recon_id_1, rel.recon_id_2 = (recon_id, t2_id)

        elif rel_type.target_entrytype is entry_table:
            # Relation[T2, T1]
            for t2_id, rel in zip(t2_ids, rels):
                rel: Relation[T2, T1]
                rel.recon_id_1, rel.recon_id_2 = (t2_id, recon_id)

        else:
            raise ValueError()

        # Add relations by recon ids
        self.add_relations_by_recon_ids(
            [(rel.recon_id_1, rel.recon_id_2, rel) for rel in rels]
        )

        # Return recon_ids assigned by db
        return t2_ids

    # Utility functions
    # ----------------------------------------------------------------------------------------------------------



# """ Contains methods to create entries to look up in the databases.
# """
#
#
# from chemrecon.core.id_types import (
#     IdentifierType, IdentifierTypeCompound, IdentifierTypeReaction, IdentifierTypeStructureRepresentation,
#     IdentifierTypeEnzyme, E_EC, IdentifierTypeAAM
# )
# from chemrecon.schema import (
#     Entry, Compound, Enzyme, Reaction, MolStructureRepr, MolStructure,
#     AAMRepr, AAM
# )
#
#
# # Direct creation of prototype entries
# # ----------------------------------------------------------------------------------------------------------------------
# def entry(id_type: IdentifierType, source_id: str) -> Entry:
#     """ Create a 'prototype' entry, which may or may not correspond to an actual entry in the database.
#     """
#     match id_type:
#         case IdentifierTypeCompound():
#             return compound_entry(id_type, source_id)
#         case IdentifierTypeReaction():
#             return reaction_entry(id_type, source_id)
#         case IdentifierTypeEnzyme():
#             return enzyme_entry(id_type, source_id)
#         case IdentifierTypeStructureRepresentation():
#             return structure_representation_entry(id_type, source_id)
#         case IdentifierTypeAAM():
#             return aam_representation_entry(id_type, source_id)
#         case _:
#             # Not implemented
#             raise NotImplementedError()
#
# def compound_entry(id_type: IdentifierTypeCompound, source_id: str) -> Compound:
#     """ TODO docs
#     """
#     return Compound(id_type = id_type, source_id = id_type.std_identifier(source_id))
#
# def reaction_entry(id_type: IdentifierTypeReaction, source_id: str) -> Reaction:
#     """ TODO docs
#     """
#     return Reaction(id_type = id_type, source_id = id_type.std_identifier(source_id))
#
# def enzyme_entry(id_type: IdentifierTypeEnzyme, source_id: str) -> Enzyme:
#     """ TODO docs
#     """
#     return Enzyme(id_type = id_type, source_id = id_type.std_identifier(source_id))
#
# def structure_representation_entry(id_type: IdentifierTypeStructureRepresentation, source_id: str) -> MolStructureRepr:
#     """ TODO docs
#     """
#     return MolStructureRepr(id_type = id_type, source_id = id_type.std_identifier(source_id))
#
# def aam_representation_entry(id_type: IdentifierTypeAAM, source_id: str) -> AAMRepr:
#     """ TODO docs
#     """
#     return AAMRepr(id_type = id_type, source_id = id_type.std_identifier(source_id))
#
# def structure_entry(smiles: str) -> MolStructure:
#     """ TODO docs
#     """
#     return MolStructure(smiles = smiles)
#
# def aam_entry(reaction_smiles: str) -> AAM:
#     """ TODO docs
#     """
#     return AAM(reaction_smiles = reaction_smiles)
#
#
# # Special for ec
# def enzyme_from_ec_number(ec_number: str) -> Enzyme:
#     """ Create a prototype enzyme entry from an EC number.
#     """
#     return Enzyme(id_type = E_EC, source_id = E_EC.std_identifier(ec_number))
#
#
# # Creation from identifiers.org strings
# # ----------------------------------------------------------------------------------------------------------------------
# def entry_from_identifiers_org(identifiers_org_string: str) -> Entry:
#     """ Create an entry from an identifiers.org string. If the string is not valid for any types in the database, raise
#         ValueError.
#     """
#     # TODO
#     # TODO raise ValueError if invalid string
#     raise NotImplementedError()



""" Enums and corresponding objects
"""

from __future__ import annotations

from enum import Enum

from chemrecon.core import id_types as id_types
from chemrecon.core.id_types import (
    IdentifierType,
)


class SourceDatabase(Enum):
    unknown = 0
    BIGG = 1
    CHEBI = 2
    ECMDB = 3
    MCSA = 4
    METAMDB = 5
    METANETX = 6
    PUBCHEM = 7
    MANUALLY_ADDED = 8
    AUTOMATIC = 9

    def __str__(self):
        return self.name
    def __repr__(self):
        return self.name


class Quality(Enum):
    unknown = 0
    automatic = 1
    preliminary = 2
    manual = 3
    manual_curated = 4

    def __str__(self):
        return self.name
    def __repr__(self):
        return self.name


class FeatureEnum(Enum):
    F = 1
    I = 2
    C = 3
    T = 4
    S = 5

    def __str__(self):
        return self.name
    def __repr__(self):
        return self.name


class IdTypeEnum(Enum):
    pass

    def __str__(self):
        return self.name
    def __repr__(self):
        return self.name


class IdTypeCompoundEnum(IdTypeEnum):
    unknown = id_types.C_UNKNOWN
    cname = id_types.C_NAME
    mnx = id_types.C_MNX
    bigg = id_types.C_BIGG
    chebi = id_types.C_CHEBI
    pubchem_cid = id_types.C_PUBCHEM
    kegg = id_types.C_KEGG
    ecmdb = id_types.C_ECMDB
    inchikey = id_types.C_INCHIKEY
    slm = id_types.C_SLM
    envipath = id_types.C_ENVIPATH
    lipidmaps = id_types.C_LIPIDMAPS
    hmdb = id_types.C_HMDB
    metacyc = id_types.C_METACYC
    seed = id_types.C_SEED
    sabiork = id_types.C_SABIORK
    reactome = id_types.C_REACTOME
    pdbe = id_types.C_PDBE
    biocyc = id_types.C_BIOCYC
    metamdb = id_types.C_METAMDB
    brenda = id_types.C_BRENDA

class IdTypeStructureRepresentationEnum(IdTypeEnum):
    unknown = id_types.S_UNKNOWN
    smiles = id_types.S_SMILES
    inchi = id_types.S_INCHI
    molfile = id_types.S_MOLFILE
    gml = id_types.S_GML


class IdTypeReactionEnum(IdTypeEnum):
    unknown = id_types.R_UNKNOWN
    rname = id_types.R_NAME
    mnx = id_types.R_MNX
    metacyc = id_types.R_METACYC
    bigg = id_types.R_BIGG
    seed = id_types.R_SEED
    kegg = id_types.R_KEGG
    rhea = id_types.R_RHEA
    sabiork = id_types.R_SABIORK
    metamdb = id_types.R_METAMDB
    mcsa = id_types.R_MCSA
    brenda = id_types.R_BRENDA


class IdTypeAAMEnum(IdTypeEnum):
    unknown = id_types.A_UNKNOWN
    reactionsmiles = id_types.A_REACTIONSMILES
    rxn = id_types.A_RXN
    gml_rule = id_types.A_GML_RULE


class IdTypeEnzymeEnum(IdTypeEnum):
    unknown = id_types.E_UNKNOWN
    ename = id_types.E_NAME
    ec = id_types.E_EC



# Enum registration
# ----------------------------------------------------------------------------------------------------------------
enum_register = [
    SourceDatabase, Quality, FeatureEnum,
    IdTypeCompoundEnum, IdTypeStructureRepresentationEnum,
    IdTypeReactionEnum, IdTypeAAMEnum, IdTypeEnzymeEnum
]

# Set the corresponding enum type for each IdType
for member in IdTypeCompoundEnum:
    if isinstance(member.value, IdentifierType):
        member.value.enum_type = IdTypeCompoundEnum.__getitem__(member.name)

for member in IdTypeReactionEnum:
    if isinstance(member.value, IdentifierType):
        member.value.enum_type = IdTypeReactionEnum.__getitem__(member.name)

for member in IdTypeStructureRepresentationEnum:
    if isinstance(member.value, IdentifierType):
        member.value.enum_type = IdTypeStructureRepresentationEnum.__getitem__(member.name)

for member in IdTypeEnzymeEnum:
    if isinstance(member.value, IdentifierType):
        member.value.enum_type = IdTypeEnzymeEnum.__getitem__(member.name)

for member in IdTypeAAMEnum:
    if isinstance(member.value, IdentifierType):
        member.value.enum_type = IdTypeAAMEnum.__getitem__(member.name)



from collections import defaultdict
from typing import Any, DefaultDict, Generator, Optional, Sequence

import psycopg as pg
import psycopg.sql as sql
from psycopg.rows import class_row

from chemrecon.schema.db_object import Column, DatabaseObject, Entry, Relation
import chemrecon.schema as schema


class QueryHandler():
    """ Context manager to wrap database queries.
    """
    conn: pg.Connection
    c: pg.Cursor

    # The handler also keeps track of a parallel 'database' of procedurally generated entries
    procedural_entries: dict[
        type[Entry], set[Entry]
    ]
    procedural_entries_index: dict[
        tuple[type[Entry], tuple],
        Entry
    ]
    procedural_entry_recon_id: dict[
        tuple[type[Entry], int],
        Entry
    ]
    procedural_id_table: dict[type[Entry], int]  # For assigning procedural ids to entries created locally


    # Context manager
    # ----------------------------------------------------------------------------------------------------------
    def __init__(self, connection: pg.Connection):
        self.conn = connection
        self.c = self.conn.cursor()

        self.procedural_entries = DefaultDict(set)
        self.procedural_entries_index = dict()
        self.procedural_entry_recon_id = dict()
        self.procedural_id_table = {
            e_type: -1 # Start at -1 and decrement
            for e_type in schema.entrytypes
        }

    def __enter__(self):
        # Enter, returning the handler as the interface
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # On leaving the context, execute the commands.
        if exc_type is not None:
            # Exception was raised in context
            raise exc_val
        self.conn.commit()
        self.c.close()

    # Procedural entries
    # ------------------------------------------------------------------------------------------------------------------
    def add_procedural_entry[T: Entry](self, entry: T) -> int:
        """ Adds a procedural entry, and returns its virtual recon id. If already exists,
            return the virtual recon id.
        """
        # TODO does the fact that index is not an ordereddict cause problems?
        index = tuple(val for val in entry.get_index_columns_with_values().values())
        try:
            # Try to return existing
            existing_entry = self.procedural_entries_index[
                type(entry), index
            ]
            entry.recon_id = existing_entry.recon_id
            return existing_entry.recon_id
        except KeyError:
            # Not found, create and insert
            new_id = self._assign_procedural_id(entry)
            self.procedural_entries[type(entry)].add(entry)
            self.procedural_entries_index[(type(entry), index)] = entry
            self.procedural_entry_recon_id[(type(entry), new_id)] = entry
            return new_id

    def _assign_procedural_id(self, entry: Entry) -> int:
        """ Assign a local id to an object, returning the id.
        """
        new_id = self.procedural_id_table[type(entry)]
        entry.recon_id = new_id
        self.procedural_id_table[type(entry)] -= 1
        return new_id

    # Row factories for Psycopg3
    # ------------------------------------------------------------------------------------------------------------------
    # Row factory for relation views
    def make_relation_entry_view_row_factory[T1: Entry, T2: Entry](
        self,
        relationtype: type[Relation[T1, T2]]
    ) -> pg.rows.RowFactory[tuple[Relation[T1, T2], T1, T2]]:

        class RelationIteratorRowFactory[T1_: Entry, T2_: Entry](pg.rows.RowFactory):
            """ Creates a row factory which constructs rows based on the relation view with attached entries.
            """

            def __init__(self, cursor: pg.Cursor[Any]):
                self.fields = [c.name for c in cursor.description]

                # Create fields dict (for fields of each table to the fields of the view)
                self.rel_fields: dict[str, str] = dict()
                self.t1_fields = dict()
                self.t2_fields = dict()

                for field in self.fields:
                    if field == 'recon_id_1':
                        self.rel_fields[field] = 'recon_id_1'
                        self.t1_fields[field] = 'recon_id'
                    elif field == 'recon_id_2':
                        self.rel_fields[field] = 'recon_id_2'
                        self.t2_fields[field] = 'recon_id'
                    elif field.startswith('rel_'):
                        self.rel_fields[field] = field.removeprefix('rel_')
                    elif field.startswith('t1_'):
                        self.t1_fields[field] = field.removeprefix('t1_')
                    elif field.startswith('t2_'):
                        self.t2_fields[field] = field.removeprefix('t2_')

            def __call__(self, values: Sequence[Any]) -> tuple[Relation[T1_, T2_], T1_, T2_]:
                # Get fields of each object
                rel_args: dict[str, Any] = dict()
                t1_args: dict[str, Any] = dict()
                t2_args: dict[str, Any] = dict()
                for fieldname, val in zip(self.fields, values):
                    # Assign values to fields
                    if fieldname in self.rel_fields:
                        rel_args[self.rel_fields[fieldname]] = val
                    if fieldname in self.t1_fields:
                        t1_args[self.t1_fields[fieldname]] = val
                    if fieldname in self.t2_fields:
                        t2_args[self.t2_fields[fieldname]] = val

                # Call constructors with the the given args
                return (
                    relationtype(**rel_args),
                    relationtype.source_entrytype(**t1_args),
                    relationtype.target_entrytype(**t2_args)
                )

        return RelationIteratorRowFactory

    # Getters
    # ----------------------------------------------------------------------------------------------------------
    def get_entry_by_recon_id[T: Entry](self, table: type[T], recon_id: int) -> T:
        """ Raises KeyError on failure. """
        res = self.get_entries_by_recon_ids(table, [recon_id])
        return res[recon_id]


    def get_entries_by_recon_ids[T: Entry](self, table: type[T], recon_ids: list[int]) -> dict[int, T]:
        """ Get the entries associated with the given recon ids. Raises KeyError on failure. """
        c = self.conn.cursor(row_factory = class_row(table))

        q = sql.SQL("""
                    SELECT * 
                    FROM {table} 
                    WHERE recon_id = %s;
        """).format(
            table = sql.Identifier(table.get_table_name())
        )
        c.executemany(
            q,
            params_seq = [[i] for i in recon_ids],
            returning = True
        )
        cursor_results = self._extract_results_from_cursor(c)
        return {
            i: e
            for i, e in zip(recon_ids, cursor_results)
        }


    def get_entry_by_index[T: Entry](self, entry_index: T) -> T:
        """ Given an entry with a valid index, get the corresponding full entry from the database.
            Raise KeyError in failure.
        """
        return self.get_entries_by_indices([entry_index])[0]


    def get_entries_by_indices[T: Entry](self, entry_indices: list[T]) -> list[T]:
        """ Given an entry with a valid index, get the corresponding full entry from the database.
            Raise KeyError in failure.
        """
        table = type(entry_indices[0])
        index_cols: list[Column] = table.get_index_columns()
        c = self.conn.cursor(row_factory = class_row(table))

        q = sql.SQL(""" 
            SELECT * 
            FROM {table}
            WHERE ({index_cols}) = ({index_placeholders})
        """).format(
            table = sql.Identifier(table.get_table_name()),
            index_cols = sql.SQL(', ').join(
                sql.Identifier(c.name) for c in index_cols
            ),
            index_placeholders = sql.SQL(', ').join(sql.SQL('%s') for _ in index_cols)
        )

        # Execute and fetch
        c.executemany(
            q,
            params_seq = [
                list(entry_index.get_index_columns_with_values().values())
                for entry_index in entry_indices
            ],
            returning = True
        )
        cursor_results = self._extract_results_from_cursor(c)
        return cursor_results


    # Relation getters
    # ----------------------------------------------------------------------------------------------------------
    def get_relations_by_recon_ids[T1: Entry, T2: Entry](
        self,
        entry_type: type[T1],
        recon_ids: list[int],
        relation_type: type[Relation[T1, T2]] | type[Relation[T2, T1]]
    ) -> list[list[Relation[T1,T2]]] | list[list[Relation[T2, T1]]]:
        """ Given a list of entries, get all connected relations of a given type.
        """
        # TODO dispatch to the three sub-functions.
        raise NotImplementedError()

    def get_relations_by_recon_ids_symmetric[T: Entry](
        self,
        recon_ids: list[int],
        relation_type: type[Relation[T, T]]
    ) -> list[list[Relation[T, T]]]:
        # Symmetric - either recon_id 1 or 2 can be found
        c = self.conn.cursor(row_factory = self.make_relation_entry_view_row_factory(relation_type))
        q = sql.SQL("""
            SELECT *
            FROM {table}
            WHERE (recon_id_1 = {recon_id_placeholder}) OR (recon_id_2 = {recon_id_placeholder});
        """).format(
            table = sql.Identifier(relation_type.get_table_name()),
            recon_id_placeholder = sql.SQL('%s')
        )
        c.executemany(q, params_seq = [[recon_id] * 2 for recon_id in recon_ids], returning = True)

        # Each input recon_id corresponds to a result set
        relations: list[list[Relation[T, T]]] = list()
        for _ in recon_ids:
            relations.append(c.fetchall())
            c.nextset()

        return relations

    def get_relations_by_recon_ids_of_t1[T1: Entry, T2: Entry](
        self,
        recon_ids: list[int],
        relation_type: type[Relation[T1, T2]]
    ) -> list[list[Relation[T1, T2]]]:
        c = self.conn.cursor(row_factory = class_row(relation_type))
        q = sql.SQL("""
                    SELECT *
                    FROM {table}
                    WHERE (recon_id_1 = {recon_id_placeholder});
                    """).format(
            table = sql.Identifier(relation_type.get_table_name()),
            recon_id_placeholder = sql.SQL('%s')
        )
        c.executemany(q, params_seq = [[recon_id] for recon_id in recon_ids], returning = True)

        # Each input recon_id corresponds to a result set
        relations: list[list[Relation[T1, T2]]] = list()
        for _ in recon_ids:
            relations.append(c.fetchall())
            c.nextset()

        return relations

    def get_relations_by_recon_ids_of_t2[T1: Entry, T2: Entry](
        self,
        recon_ids: list[int],
        relation_type: type[Relation[T2, T1]]
    ) -> list[list[Relation[T2, T1]]]:
        c = self.conn.cursor(row_factory = class_row(relation_type))
        q = sql.SQL("""
                    SELECT *
                    FROM {table}
                    WHERE (recon_id_2 = {recon_id_placeholder});
        """).format(
            table = sql.Identifier(relation_type.get_table_name()),
            recon_id_placeholder = sql.SQL('%s')
        )
        c.executemany(q, params_seq = [[recon_id] for recon_id in recon_ids], returning = True)

        # Each input recon_id corresponds to a result set
        relations: list[list[Relation[T1, T2]]] = list()
        for _ in recon_ids:
            relations.append(c.fetchall())
            c.nextset()

        return relations


    # Relation and entry getters
    # ----------------------------------------------------------------------------------------------------------
    def get_relations_with_entries_by_recon_ids[T1: Entry, T2: Entry](
        self,
        entry_type: type[T1],
        recon_ids: list[int],
        relation_type: type[Relation[T1, T2]] | type[Relation[T2, T1]],
    ) -> list[dict[T2, list[Relation[T1, T2]] | list[Relation[T2, T1]]]]:
        """ Given a list of entries, get connected relations of the given type.
            Returns a list, containing, for each input recon_id, a dictionary mapping connected vertices
            to the relation(s) with which they are connected.
        """
        # First, establish the types
        t2_entrytype = (
            relation_type.source_entrytype if relation_type.target_entrytype is entry_type
            else relation_type.target_entrytype
        )
        is_symmetric = (entry_type is t2_entrytype)

        # Dispatch to sub-functions depending on types
        if is_symmetric:
            # Case 1: Symmetric - either recon_id 1 or 2 can be found
            return self.get_relations_with_entries_by_recon_ids_symmetric(entry_type, recon_ids, relation_type)
        else:
            if relation_type.source_entrytype is entry_type:
                # Case 2: Relation is [T1, T2]
                return self.get_relations_with_entries_by_recon_ids_of_t1(entry_type, recon_ids, relation_type)
            elif relation_type.target_entrytype is entry_type:
                # Case 3: Relation is [T2, T1]
                return self.get_relations_with_entries_by_recon_ids_of_t2(entry_type, recon_ids, relation_type)
            else:
                # Invalid shape of entrytype and relation?
                raise ValueError()

    # Specialized getters for different entry/relation cases
    def get_relations_with_entries_by_recon_ids_symmetric[T: Entry](
        self,
        entry_type: type[T],
        recon_ids: list[int],
        relation_type: type[Relation[T, T]],
    ) -> list[dict[T, list[Relation[T, T]]]]:
        """ Specialization for symmetric relations.
        """
        c = self.conn.cursor(row_factory = class_row(relation_type))

        # TODO REDO, call above functions
        # TODO get by views!

        # Symmetric - either recon_id 1 or 2 can be found
        q = sql.SQL("""
                    SELECT *
                    FROM {table}
                    WHERE (recon_id_1 = {recon_id_placeholder}) OR (recon_id_2 = {recon_id_placeholder});
                    """).format(
            table = sql.Identifier(relation_type.get_table_name()),
            recon_id_placeholder = sql.SQL('%s')
        )
        c.executemany(q, params_seq = [[recon_id] * 2 for recon_id in recon_ids], returning = True)

        # Each input recon_id corresponds to a result set
        entry_relations: dict[int, list[Relation]] = dict()
        for recon_id in recon_ids:
            entry_relations[recon_id] = c.fetchall()
            c.nextset()

        # Get entries
        results: list[dict[T, list[Relation[T, T]]]] = list()

        for input_recon_id, rels in entry_relations.items():
            # For each source recon id, run a query to get relation targets
            subresults: dict[T, list[Relation[T, T]]] = defaultdict(list)
            target_recon_ids = [  # List corresponding to the list of relations rels
                rel.recon_id_1 if rel.recon_id_2 == input_recon_id else rel.recon_id_2
                for rel in rels
            ]
            q = sql.SQL("""
                        SELECT *
                        FROM {t2_table}
                        WHERE recon_id = %s;
                        """).format(
                t2_table = sql.Identifier(entry_type.get_table_name())
            )
            c = self.conn.cursor(row_factory = class_row(entry_type))
            c.executemany(q, params_seq = [[tid] for tid in target_recon_ids], returning = True)
            target_entries: list[T] = self._extract_results_from_cursor(c)

            # Add to results
            for rel, target_entry in zip(rels, target_entries):
                subresults[target_entry].append(rel)
            results.append(subresults)

        # Return
        return results

    def get_relations_with_entries_by_recon_ids_of_t1[T1: Entry, T2: Entry](
        self,
        entry_type: type[T1],
        recon_ids: list[int],
        relation_type: type[Relation[T1, T2]],
    ) -> list[dict[T2, list[Relation[T1, T2]]]]:
        """ Given a list of entries, get connected relations of the given type.
            Returns a list, containing, for each input recon_id, a dictionary mapping connected vertices
            to the relation(s) with which they are connected.
        """
        c = self.conn.cursor(row_factory = self.make_relation_entry_view_row_factory(relation_type))
        q = sql.SQL("""
                    SELECT *
                    FROM {view}
                    WHERE (recon_id_1 = {recon_id_placeholder});
                    """).format(
            view = sql.Identifier(f'{relation_type.get_table_name()}_v'),
            recon_id_placeholder = sql.SQL('%s')
        )
        c.executemany(q, params_seq = [[recon_id] for recon_id in recon_ids], returning = True)

        # Process output
        result: list[dict[T2, list[Relation[T1, T2]]]] = list()
        cursor_result = self._extract_results_from_cursor_by_sets(c)
        for recon_id, relation_list in zip(recon_ids, cursor_result):
            subdict: dict[T2, list[Relation[T1, T2]]] = defaultdict(list)
            for relation, entry_1, entry_2 in relation_list:
                assert entry_1.recon_id == recon_id
                subdict[entry_2].append(relation)
            result.append(subdict)
        return result

    def get_relations_with_entries_by_recon_ids_of_t2[T1: Entry, T2: Entry](
        self,
        entry_type: type[T1],
        recon_ids: list[int],
        relation_type: type[Relation[T2, T1]],
    ) -> list[dict[T2, list[Relation[T2, T1]]]]:
        """ Given a list of entries, get connected relations of the given type.
            Returns a list, containing, for each input recon_id, a dictionary mapping connected vertices
            to the relation(s) with which they are connected.
        """
        c = self.conn.cursor(row_factory = self.make_relation_entry_view_row_factory(relation_type))
        q = sql.SQL("""
                    SELECT *
                    FROM {view}
                    WHERE (recon_id_2 = {recon_id_placeholder});
                    """).format(
            view = sql.Identifier(f'{relation_type.get_table_name()}_v'),
            recon_id_placeholder = sql.SQL('%s')
        )
        c.executemany(q, params_seq = [[recon_id] for recon_id in recon_ids], returning = True)

        # Process output
        result: list[dict[T2, list[Relation[T2, T1]]]] = list()
        cursor_result = self._extract_results_from_cursor_by_sets(c)
        for recon_id, relation_list in zip(recon_ids, cursor_result):
            subdict: dict[T2, list[Relation[T2, T1]]] = defaultdict(list)
            for relation, entry_2, entry_1 in relation_list:
                assert entry_1.recon_id == recon_id
                subdict[entry_2].append(relation)
            result.append(subdict)
        return result

    # Graph exploration getters
    # ----------------------------------------------------------------------------------------------------------
    # These are the main queries used to efficiently explore the graph.
    def explore_vertex[T: Entry](
        self,
        entry: T,
        relation_types: type[Relation[T, Any]]
    ) -> list[tuple[Entry, list[Relation]]]:
        """ Explores the neighbouring entries to a given entry.
            Results are returned as a list of tuples, each containing a neighbouring entry and the relations
            which point to it.
            Note that the input entry may be given in the output in the case of reflexive relations.
        """

        # TODO Run get_relations_with_entries_from_source() for each allowed relation, and collate the results

        # TODO
        raise NotImplementedError()

    # Misc
    # ----------------------------------------------------------------------------------------------------------
    def get_number_of_entries[T: DatabaseObject](self, table: type[T]) -> int:
        """ Return hte number of entries in the table"""
        table_name = table.get_table_name()
        q = sql.SQL("""SELECT count(*) FROM {table}""").format(
            table = sql.Identifier(table_name)
        )
        self.c.execute(q)
        result: int = self.c.fetchone()[0]
        return result

    def iterate_table[T: DatabaseObject](
        self,
        table: type[T],
        batch_size: int = 1000,
        up_to: int = 0
    ) -> Generator[T, None, None]:
        """ Creates an iterator which queries all entries in a table.
            Entries are requested from the database in batches, but returned as a continuous iterator.
        """
        i: int = 0
        batch: list[T] = list()

        with self.conn.cursor(row_factory = class_row(table)) as cursor:
            q = sql.SQL("""
                SELECT *
                FROM {table}
                ;
            """).format(
                table = sql.Identifier(table.get_table_name())
            )
            try:
                cursor.execute(q)
            except Exception as e:
                raise RuntimeError('!: {e}')

            while True:
                # Main loop, fetch batches and yield from the batches
                batch = cursor.fetchmany(size = batch_size)
                if not batch:
                    break
                item: T
                for item in batch:
                    i += 1
                    yield item
                    if up_to and i > up_to:
                        return


    def iterate_relation_with_entries[T1: Entry, T2: Entry](
        self,
        relationtype: type[Relation[T1, T2]],
        batch_size: int = 1000,
        up_to: int = 0
    ) -> Generator[tuple[Relation[T1, T2], T1, T2], None, None]:
        """ Iterates the view over a relation table, getting entries adjacent to each relation.
        """
        i: int = 0
        batch: list = list()

        # TODO make a relationIteratorRowFactory and use

        # With the factory created, iterate the rows
        with self.conn.cursor(row_factory = self.make_relation_entry_view_row_factory(relationtype)) as cursor:
            q = sql.SQL("""
                        SELECT *
                        FROM {view}
                        ;
            """).format(
                view = sql.Identifier(f'{relationtype.get_table_name()}_v')
            )
            try:
                cursor.execute(q)
            except Exception as e:
                raise RuntimeError('!: {e}')

            while True:
                # Main loop, fetch batches and yield from the batches
                batch = cursor.fetchmany(size = batch_size)
                if not batch:
                    break
                item: tuple[Relation[T1, T2], T1, T2]
                for item in batch:
                    i += 1
                    yield item
                    if up_to and i > up_to:
                        return

    def _extract_results_from_cursor(self, cursor: Optional[pg.Cursor] = None) -> list[Any]:
        """ Iterates over all result sets of the Psycopg3 cursor and returns.
        """
        c = cursor or self.c
        result: list[Any] = [c.fetchone()]
        while c.nextset():
            result.append(c.fetchone())
        return result

    def _extract_results_from_cursor_by_sets(self, cursor: Optional[pg.Cursor] = None) -> list[list[Any]]:
        """ Iterates over all result sets of the Psycopg3 cursor and returns.
        """
        c = cursor or self.c
        result: list[list[Any]] = [c.fetchall()]
        while c.nextset():
            result.append(c.fetchall())
        return result



""" Defines procedural entries which are generated using entry graphs.
"""
from __future__ import annotations

from abc import ABC
from typing import ClassVar, final, Optional

import chemrecon.entrygraph.scoring as scoring

import chemrecon.entrygraph.explore as explore
from chemrecon.schema.db_object import Entry, ProceduralRelation
from chemrecon.schema.db_object import col_score
from chemrecon.entrygraph.explorationprotocol import ExplorationProtocol
from chemrecon.entrygraph.entrygraph import EntryGraph


class ProceduralRelationEG[T1: Entry, T2: Entry](ProceduralRelation[T1, T2], ABC):
    """ A procedural relation which is calculated based on an entrygraph of other relations
        in the database.
        Requires entrygraph to only accept entries of type T1 as initial entries, and the
        scorer to score entries of type T2.
    """
    # Attribute
    score: float  #: The score of the target entry in the entry graph.

    # Defines the parameters from which the EG is generated
    protocol: ClassVar[ExplorationProtocol]
    entrygraph_scorer: ClassVar[scoring.Scorer]

    # Necessary
    symmetric = False
    _attribute_columns = [
        col_score
    ]
    _index = []


    def __init__(
        self,
        recon_id_1: Optional[int] = None,
        recon_id_2: Optional[int] = None,
        score: float = 0.0
    ):
        self.score = score
        super().__init__(recon_id_1 = recon_id_1, recon_id_2 = recon_id_2)

    @final
    @classmethod
    def generate(
        cls,
        take_entry: T1,
    ) -> list[tuple[ProceduralRelation[T1, T2], T2]]:
        """ Generation of procedural relations relies on generating the associated entrygraph.
        """
        eg = EntryGraph(
            initial_entries = {take_entry}
        )
        explore.explore(eg, steps = 4, protocol = cls.protocol)
        score = cls.entrygraph_scorer(eg)
        return [
            (cls(recon_id_1 = None, recon_id_2 = None, score = score), score_entry)
            for score_entry, score in score.items()
        ]

    def _vis_str(self) -> str:
        return f'score: {self.score}'



from typing import Optional

from chemrecon.schema.db_object import col_src
from chemrecon.schema.enums import SourceDatabase
from chemrecon.schema import Compound, MolStructure, ComposedRelation
from chemrecon.schema.entry_types.molstructure_repr import MolStructureRepr
from chemrecon.schema.relation_types_source.compound_has_structure_representation_relation import (
    CompoundHasStructureRepresentation
)
from chemrecon.schema.relation_types_procedural.molstructure_convert_relation import MolStructureConvert

class CompoundHasMolStructure(ComposedRelation[Compound, MolStructure, MolStructureRepr]):
    """ This relation gives the associated molecular structures of a compound in a standardized format.
        This relation is composed of the `CompoundHasMolStructureRepr` relation and the `MolStructureConvert`
        relations.
        So if a compound entry includes an InChI, Molfile representation of a structure, this relation gives the same
        structure in the standardized, SMILES-based format.
    """
    # Attributes
    src: SourceDatabase  #: The source database containing this structure.

    # Transitive
    rel_type_1 = CompoundHasStructureRepresentation
    rel_type_2 = MolStructureConvert
    intermediate_entrytype = MolStructureRepr

    # Database
    entrytype_name = 'Compound has MolStructure'
    _table_name = 'CompoundHasMolStructure'
    symmetric = False
    source_entrytype = Compound
    target_entrytype = MolStructure
    _attribute_columns = [
        col_src
    ]
    _index = [0]

    # Filters
    @classmethod
    def filter_intermediate(cls, e: MolStructureRepr) -> bool:
        if e.implicit:
            return False
        return True

    def __init__(
            self,
            rel_1: CompoundHasStructureRepresentation,
            rel_2: MolStructureConvert,
            intermediate: MolStructureRepr,
            recon_id_1: Optional[int] = None,
            recon_id_2: Optional[int] = None
    ):
        # Inherits src from rel1
        self.src = rel_1.src
        super().__init__(rel_1, rel_2, intermediate, recon_id_1, recon_id_2)

    def _vis_str(self) -> str:
        return f'{self.src.name}'




from typing import Optional

from chemrecon.schema import IdTypeReactionEnum, Quality
from chemrecon.schema.db_object import (
    Column, SourceEntry, col_name, col_properties, col_quality,
    col_recon_id,
    col_source_id,
)

class Reaction(SourceEntry):
    """ Represents a compound entry. """
    # Attributes
    source_id: str  #: Identifier of the database entry. (**index**)
    id_type: IdTypeReactionEnum  #: Source of the database entry. (**index**)
    name: Optional[str]
    is_transport: Optional[bool]  #: Whether the reaction is explicitly marked as a transport reaction.
    is_reversible: Optional[bool]  #: Whether the reaction is explicitly marked as reversible.
    quality: Optional[Quality]  #: Indicates the quality of the entry, if this is specified by the source database.
    properties: list[str]  #: Additional properties

    # Database
    entrytype_name = 'Reaction'
    _table_name = 'Reaction'
    _columns = [
        col_recon_id,
        col_source_id,
        Column('id_type', IdTypeReactionEnum),
        col_name,
        Column('is_transport', bool),
        Column('is_reversible', bool),
        col_quality,
        col_properties,
    ]
    _index = [1, 2]

    # Visualisation
    _draw_colour = '9EFFA1'

    def __init__(
            self,
            source_id: str,
            id_type: IdTypeReactionEnum,
            name: Optional[str] = None,
            is_transport: Optional[bool] = None,
            is_reversible: Optional[bool] = None,
            quality: Optional[Quality] = None,
            recon_id: Optional[int] = None,
            properties: Optional[list[str]] = None,
    ):
        super().__init__(recon_id)
        self.source_id = source_id
        self.id_type = id_type
        self.name = name
        self.is_transport = is_transport
        self.is_reversible = is_reversible
        self.quality = quality
        self.properties = properties if properties else list()



""" Handles the creation of the tables described in 'core/schema.py'.
    All tables and objects are placed in the 'chemrecon' schema.
    Should ensure admin access of the connection before attempting to create tables.
    If tables already exist, verify that they are identical.
    This module has the following expectations of the database being connected to:
    - Database named 'chemrecon_db'
    - Initial login with 'postgres' user
    - Afterwards, login with either local_docker or local_docker_dev
    -
    -
    -
    -
"""
import argparse
import builtins
import os
from enum import Enum
from typing import get_origin, get_args
from types import GenericAlias

import psycopg as pg
import psycopg.errors
import psycopg.sql as sql

import chemrecon.schema as schema
from chemrecon import Params

from chemrecon.database import params
from chemrecon.database.connect import postgres_connect
from chemrecon.utils.set_cwd import set_cwd

from chemrecon.schema import entrytypes, relationtypes, InverseRelation, Relation

# Command-line-arguments
parser = argparse.ArgumentParser()
# Whether to set up local or public users
parser.add_argument('--params', action = 'store')
parser.add_argument('--devparams', action = 'store')

# Custom params
parser.add_argument('--host', action = 'store')
parser.add_argument('--port', action = 'store')
parser.add_argument('--database', action = 'store')
parser.add_argument('--username', action = 'store')
parser.add_argument('--password', action = 'store')


def main():
    print(f'Running ChemRecon database initialisation script.')
    args = vars(parser.parse_args())

    print(f'\nARGS\n')
    print(args)

    if 'params' not in args:
        raise ValueError('Must provide parameters file')
    if 'devparams' not in args:
        raise ValueError('Must provide dev parameters file.')

    # Connection params
    connection_params = Params(
        connection_title = 'CLI',
        db_name = args['database'],
        db_host = args['host'],
        db_port = args['port'],
        username = args['username'],
        password = args['password']
    )

    print(os.getcwd())

    # User parameters
    with open(args['params']) as f:
        lines = f.readlines()
        params_user = Params(
            connection_title = 'CLI-user',
            db_name = lines[0].strip(),
            username = lines[1].strip(),
            password = lines[2].strip(),
            db_host = lines[3].strip(),
            db_port = lines[4].strip()
        )
    with open(args['devparams']) as f:
        lines = f.readlines()
        params_dev = Params(
            connection_title = 'CLI-user',
            db_name = lines[0].strip(),
            username = lines[1].strip(),
            password = lines[2].strip(),
            db_host = lines[3].strip(),
            db_port = lines[4].strip()
        )

    conn = postgres_connect(
        connection_params,
        initialise_enums = False
    )

    # Initialize
    initialise_database(
        conn,
        dev_parameters = params_dev,
        pub_parameters = params_user,
    )


def initialise_database(
        conn: pg.Connection,
        dev_parameters: Params,
        pub_parameters: Params,
        only_create_users: bool = False
):
    # Login with the 'postgres' user to create the DB.
    # Create the database, schema and accounts from an otherwise empty database

    # Get connection parameters
    set_cwd()

    print(f'\nCreating database...')

    with conn.cursor() as c:

        # Check if DB is already populated
        # TODO

        # Load parameters
        # ------------------------------------------------------------------------------------------------------
        db_user_dev = dev_parameters.username
        db_pass_dev = dev_parameters.password

        db_user_pub = pub_parameters.username
        db_pass_pub = pub_parameters.password

        # Create users if not exist
        try:

            # Create group for public users
            # ----------------------------------------------------------------------------------------------------------

            c.execute(sql.SQL(
                'CREATE GROUP {public_group};'
            ).format(
                public_group = sql.Identifier('public_group')
            ))

            # Create users/roles
            # ----------------------------------------------------------------------------------------------------------
            c.execute(sql.SQL(
                'CREATE USER {dev_username} WITH PASSWORD {dev_password};'
            ).format(
                dev_username = sql.Identifier(db_user_dev),
                dev_password = db_pass_dev,
            ))
            c.execute(sql.SQL(
                'CREATE USER {pub_username} WITH PASSWORD {pub_password} IN GROUP {public_group};'
            ).format(
                pub_username = sql.Identifier(db_user_pub),
                pub_password = db_pass_pub,
                public_group = sql.Identifier('public_group')
            ))

        except psycopg.errors.DuplicateObject as e:
            # Already exists, reset
            conn.commit()
            print(f' -> Users already exist, skipping.')
            pass

        # If set to only prepare for restoring with psql, exit here
        # --------------------------------------------------------------------------------------------------------------
        if only_create_users:
            conn.commit()
            return

        # Create schemata
        # --------------------------------------------------------------------------------------------------------------
        c.execute(sql.SQL(  # 'meta' schema for dev purposes
            'CREATE SCHEMA {meta_schema} AUTHORIZATION {dev_username};'
        ).format(
            meta_schema = sql.Identifier('meta'),
            dev_username = sql.Identifier(db_user_dev)
        ))
        c.execute(sql.SQL(  # 'chemrecon' schema for data
            'CREATE SCHEMA {chemrecon_schema} AUTHORIZATION {dev_username};'
        ).format(
            chemrecon_schema = sql.Identifier('chemrecon'),
            dev_username = sql.Identifier(db_user_dev)
        ))

        # Create the 'meta' tables
        # --------------------------------------------------------------------------------------------------------------
        c.execute(sql.SQL("""
                          CREATE TABLE IF NOT EXISTS meta.procedures
                          (
                              procedure_name
                              TEXT,
                              procedure_version
                              INT,
                              data_source_version
                              TEXT,
                              finished_time
                              TIMESTAMP,
                              progress
                              INT
                          );
                          """))
        conn.commit()

        # Create and adapt enums
        # --------------------------------------------------------------------------------------------------------------
        print(f'Creating enums')
        for enum in schema.enums.enum_register:
            # Create enum in db
            c.execute(sql.SQL(
                'CREATE TYPE {chemrecon}.{enum_name} AS ENUM ({enum_list});'
            ).format(
                chemrecon = sql.Identifier('chemrecon'),
                enum_name = sql.Identifier(enum.__name__),
                enum_list = sql.SQL(', ').join(
                    e.name for e in enum
                )
            ))
        conn.commit()

        # Create tables
        # --------------------------------------------------------------------------------------------------------------
        # Entry tables
        # --------------------------------------------------------------------------------------------------------------
        for entrytype in entrytypes:
            table = sql.Identifier(entrytype.get_table_name())
            column_sqls: list[sql.Composable] = list()

            # Compute the columns
            for i, col in enumerate(entrytype.get_columns()):
                col: schema.db_object.Column
                colname = sql.Identifier(col.name)
                coltype = get_postgres_type(col.col_type)
                col_sql = sql.SQL(
                    '{colname} {coltype} {opt_primary_key}'
                ).format(
                    colname = colname,
                    coltype = coltype if not col.serial else sql.SQL('SERIAL'),
                    opt_primary_key = sql.SQL('PRIMARY KEY') if i == 0 else sql.SQL('')
                )
                column_sqls.append(col_sql)

            # Create the table
            q = sql.SQL(
                'CREATE TABLE IF NOT EXISTS {table} ({cols});'
            ).format(
                table = table,
                cols = sql.SQL(', ').join(column_sqls),
            )
            print(f' - {q.as_string()}')
            c.execute(q)

            # Create index and unique constraint on index columns
            if len(entrytype.get_index_indices()) > 0:
                index_cols: list[sql.Composable] = list()
                for col in entrytype.get_index_columns():
                    if col.index_hash:
                        # If column has potential to be long, index the md5 value of the column instead
                        index_cols.append(sql.SQL('md5({colname})').format(colname = sql.Identifier(col.name)))
                    else:
                        index_cols.append(sql.Identifier(col.name))
                q = sql.SQL(
                    'CREATE UNIQUE INDEX IF NOT EXISTS {index_name} ON {table} ({index_cols});'
                ).format(
                    index_name = sql.Identifier(f'{entrytype.get_table_name()}_index'),
                    constraint_name = sql.Identifier(f'{entrytype.get_table_name()}_constraint'),
                    table = table,
                    index_cols = sql.SQL(', ').join(index_cols)
                )
                print(f' - {q.as_string()}')
                c.execute(q)

        # Relation Tables
        # --------------------------------------------------------------------------------------------------------------
        for relationtype in relationtypes:
            if issubclass(relationtype, InverseRelation):
                # Inverse relations are handles in the next step
                continue

            table = relationtype.get_table_name()
            column_sqls: list[sql.Composable] = get_column_sql_terms(relationtype)

            # Table query
            q = sql.SQL(
                'CREATE TABLE {rel_table_name} ({cols} {optional_check});'
            ).format(
                rel_table_name = sql.Identifier(table),
                cols = sql.SQL(', ').join(column_sqls),
                optional_check = (  # If relation is from/to the same table, ensure invariant id_1 <= id_2
                    sql.SQL(', CHECK (recon_id_1 <= recon_id_2)') if relationtype.symmetric
                    else sql.SQL('')
                )
            )
            print(f' - {q.as_string()}')
            c.execute(q)

            # Create unique constraint on (recon_id_1, recon_id_2, *index cols)
            index_cols: list[sql.Composable] = [
                sql.Identifier(col.name) for col in relationtype.get_index_columns()
            ]
            q = sql.SQL(
                'CREATE UNIQUE INDEX {index_name} ON {table} ({index_cols})'
            ).format(
                index_name = sql.Identifier(f'{relationtype.get_table_name()}_index'),
                table = sql.Identifier(relationtype.get_table_name()),
                index_cols = sql.SQL(', ').join(index_cols)
            )
            print(f' - {q.as_string()}')
            c.execute(q)

        # Create views for inverse relations
        # --------------------------------------------------------------------------------------------------------------
        for relationtype in relationtypes:
            if not issubclass(relationtype, InverseRelation):
                # Inverse relations are handles in the next step
                continue

            # Sanity checks
            assert relationtype.inverse_main_relation.target_entrytype is relationtype.source_entrytype
            assert relationtype.inverse_main_relation.source_entrytype is relationtype.target_entrytype

            # Create table as view
            table = relationtype.get_table_name()
            q = sql.SQL(""" 
                CREATE VIEW {view_name} AS
                    SELECT recon_id_1 AS recon_id_2, recon_id_2 AS recon_id_1, {attr_cols}
                    FROM {parent_table_name}
                    ;
            """).format(
                view_name = sql.Identifier(table),
                parent_table_name = sql.Identifier(relationtype.inverse_main_relation.get_table_name()),
                attr_cols = sql.SQL(', ').join(
                    sql.Identifier(attr_col.name) for attr_col in relationtype.get_attribute_columns()
                )
            )
            print(f' - {q.as_string()}')
            c.execute(q)

        # Create corresponding views for ALL relations
        # --------------------------------------------------------------------------------------------------------------
        for relationtype in relationtypes:
            # Create view corresponding to this relation (relation with adjacent entries)
            table = relationtype.get_table_name()
            q = sql.SQL("""
                        CREATE VIEW {view_name} AS
                        SELECT rel.recon_id_1, rel.recon_id_2, {rel_attr_cols} {t1_attr_cols}, {t2_attr_cols}
                        FROM {rel_table} rel
                                 INNER JOIN {table_1} t1 ON t1.recon_id = rel.recon_id_1
                                 INNER JOIN {table_2} t2 ON t2.recon_id = rel.recon_id_2
                        ;
                        """).format(
                view_name = sql.Identifier(f'{table}_v'),
                rel_table = sql.Identifier(table),
                table_1 = sql.Identifier(relationtype.source_entrytype.get_table_name()),
                table_2 = sql.Identifier(relationtype.target_entrytype.get_table_name()),
                rel_attr_cols = sql.SQL(', ').join(
                    sql.SQL('rel.') + sql.Identifier(c.name)
                    + sql.SQL(' ') + sql.Identifier(f'rel_{c.name}')
                    for c in relationtype.get_attribute_columns()
                ) + (sql.SQL(', ') if len(relationtype.get_attribute_columns()) > 0 else sql.SQL('')),
                t1_attr_cols = sql.SQL(', ').join(
                    sql.SQL('t1.') + sql.Identifier(c.name)
                    + sql.SQL(' ') + sql.Identifier(f't1_{c.name}')
                    for c in relationtype.source_entrytype.get_columns(include_recon_id = False)
                ),
                t2_attr_cols = sql.SQL(', ').join(
                    sql.SQL('t2.') + sql.Identifier(c.name)
                    + sql.SQL(' ') + sql.Identifier(f't2_{c.name}')
                    for c in relationtype.target_entrytype.get_columns(include_recon_id = False)
                ),
            )

            print(f' -> {relationtype.get_table_name()}')
            print(f'   - {q.as_string()}')
            c.execute(q)

        # Grant read/write permissions to 'dev' user on all tables
        # --------------------------------------------------------------------------------------------------------------
        for schema_name in {'meta', 'chemrecon'}:
            c.execute(sql.SQL(
                """ GRANT SELECT, INSERT, UPDATE, DELETE 
                    ON ALL TABLES IN SCHEMA {schema_name} 
                    TO {dev_username};
                    GRANT USAGE, SELECT 
                    ON ALL SEQUENCES IN SCHEMA {schema_name} 
                    TO {dev_username};
                """
            ).format(
                schema_name = sql.Identifier(schema_name),
                dev_username = sql.Identifier(db_user_dev)
            ))

        # Grant USAGE, UPDATE on sequences
        # TODO

        # Grant read permissions to 'public' group on all tables
        # --------------------------------------------------------------------------------------------------------------
        for schema_name in {'meta', 'chemrecon'}:
            c.execute(sql.SQL(
                """ GRANT USAGE ON SCHEMA {schema_name} TO {pub_username};
                    GRANT SELECT 
                    ON ALL TABLES IN SCHEMA {schema_name} 
                    TO {pub_username};
                    GRANT SELECT
                    ON ALL SEQUENCES IN SCHEMA {schema_name}
                    TO {pub_username};
                """
            ).format(
                schema_name = sql.Identifier(schema_name),
                pub_username = sql.Identifier('public_group')
            ))

        # Grant USAGE on sequences
        # TODO

        # Commit and finalise
        # --------------------------------------------------------------------------------------------------------------
        print('Database initialized.')
        conn.commit()


# Utility functions
# ----------------------------------------------------------------------------------------------------------------------
def get_column_sql_terms(relationtype: type[Relation], only_attrs: bool = False) -> list[sql.Composable]:
    """ Generate SQL statemetns to create each column
    """
    column_sqls: list[sql.Composable] = list()

    # Source and target columns
    if not only_attrs:
        column_sqls.append(sql.SQL(
            'recon_id_1 int references {table_1}(recon_id)'
        ).format(
            table_1 = sql.Identifier(relationtype.source_entrytype.get_table_name()),
        ))
        column_sqls.append(sql.SQL(
            'recon_id_2 int references {table_2}(recon_id)'
        ).format(
            table_2 = sql.Identifier(relationtype.target_entrytype.get_table_name()),
        ))

    # Attribute columns
    for attrcol in relationtype.get_attribute_columns():
        column_sqls.append(sql.SQL(
            '{colname} {coltype}'
        ).format(
            colname = sql.Identifier(attrcol.name),
            coltype = get_postgres_type(attrcol.col_type)
        ))

    return column_sqls

def get_postgres_type(t: type) -> sql.Composable:
    # Get the postgres type as an SQL string of the given type
    match t:
        case builtins.str:
            return sql.SQL('text')
        case builtins.int:
            return sql.SQL('integer')
        case builtins.float:
            return sql.SQL('float8')
        case builtins.bool:
            return sql.SQL('boolean')
        case GenericAlias():
            # Subscripted type, e.g. list[int] -> 'integer[]'
            origin_type = get_origin(t)
            match origin_type:
                case builtins.list:
                    # List gets translated into postgres array
                    subscript = get_args(t)[0]
                    subscript_type = get_postgres_type(subscript)
                    q =  sql.SQL('{subscript_sql} ARRAY').format(
                        subscript_sql = subscript_type
                    )
                    return q
                case _:
                    raise NotImplementedError('Implement creating table from whatever type this  is.')
        case _:
            # If it is an enum
            if issubclass(t, Enum):
                return sql.SQL('{chemrecon}.{enumname}').format(
                    chemrecon = sql.Identifier('chemrecon'),
                    enumname = sql.Identifier(t.__name__)
                )
            else:
                raise NotImplementedError('!')


# Script functionality
if __name__ == '__main__':
    main()



from __future__ import annotations

from enum import Enum


class Direction(Enum):
    FORWARDS = 1  #: From source to target.
    BACKWARDS = 2  #: From target to source.
    BOTH = 3  #: Both of the above.
    SYMMETRIC = 4  #: To be used for symmetric relations.



from typing import Optional

from chemrecon.chem.mol import Mol
from chemrecon.chem.create_mol import mol_from_structurerepresentation
from chemrecon.schema import MolStructureRepr, MolStructure
from chemrecon.schema.db_object import ProceduralGeneratorError, ProceduralRelation


class MolStructureConvert(ProceduralRelation[MolStructureRepr, MolStructure]):
    """ Standardized and canonical conversion of molecular structure representations (e.g. SMILES, InChI, ...) into
        a consistent format, stored as a canonical SMILES string.
    """
    # Attributes

    # Database
    entrytype_name = 'MolStructureConvert'
    _table_name = 'MolStructureConvert'
    symmetric = False
    source_entrytype = MolStructureRepr
    target_entrytype = MolStructure
    _attribute_columns = []
    _index = []

    def __init__(
        self,
        recon_id_1: Optional[int] = None,
        recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)


    @classmethod
    def generate(
        cls, take_entry: MolStructureRepr
    ) -> list[tuple[ProceduralRelation[MolStructureRepr, MolStructure], MolStructure]]:

        try:
            mol: Mol = mol_from_structurerepresentation(take_entry, template = True)
            new_entry: MolStructure = mol.to_database_struct()
            return [
                (
                    MolStructureConvert(),
                    new_entry
                )
            ]

        except Exception as e:
            # Failed for this item
            raise ProceduralGeneratorError(f'Creating molecule failed for {take_entry.source_id}') from e


""" Basic chemical information.
"""
from typing import NamedTuple

class Element(NamedTuple):
    num: int
    symbol: str
    name: str

n =     Element(0,      'n',    'Neutron')  # Included for correct list indices
H =     Element(1,      'H',    'Hydrogen')
He =    Element(2,      'He',   '')
Li =    Element(3,      'Li',   '')
Be =    Element(4,      'Be',   '')
B =     Element(5,      'B',    '')
C =     Element(6,      'C',    '')
N =     Element(7,      'N',    '')
O =     Element(8,      'O',    '')
F =     Element(9,      'F',    '')
Ne =    Element(10,     'Ne',   '')
Na =    Element(11,     'Na',   '')
Mg =    Element(12,     'Mg',   '')
Al =    Element(13,     'Al',   '')
Si =    Element(14,     'Si',   '')
P =     Element(15,     'P',    '')
S =     Element(16,     'S',    '')
Cl =    Element(17,     'Cl',   '')
Ar =    Element(18,     'Ar',   '')
K =     Element(19,     'K',    '')
Ca =    Element(20,     'Ca',   '')
Sc =    Element(21,     'Sc',   '')
Ti =    Element(22,     'Ti',   '')
V =     Element(23,     'V',    '')
Cr =    Element(24,     'Cr',   '')
Mn =    Element(25,     'Mn',   '')
Fe =    Element(26,     'Fe',   '')
Co =    Element(27,     'Co',   '')
Ni =    Element(28,     'Ni',   '')
Cu =    Element(29,     'Cu',   '')
Zn =    Element(30,     'Zn',   '')
Ga =    Element(31,     'Ga',   '')
Ge =    Element(32,     'Ge',   '')
As =    Element(33,     'As',   '')
Se =    Element(34,     'Se',   '')
Br =    Element(35,     'Br',   '')
Kr =    Element(36,     'Kr',   '')
Rb =    Element(37,     'Rb',   '')
Sr =    Element(38,     'Sr',   '')
Y =     Element(39,     'Y',    '')
Zr =    Element(40,     'Zr',   '')
Nb =    Element(41,     'Nb',   '')
Mo =    Element(42,     'Mo',   '')
Tc =    Element(43,     'Tc',   '')
Ru =    Element(44,     'Ru',   '')
Rh =    Element(45,     'Rh',   '')
Pd =    Element(46,     'Pd',   '')
Ag =    Element(47,     'Ag',   '')
Cd =    Element(48,     'Cd',   '')
In =    Element(49,     'In',   '')
Sn =    Element(50,     'Sn',   '')
Sb =    Element(51,     'Sb',   '')
Te =    Element(52,     'Te',   '')
I =     Element(53,     'I',    '')
Xe =    Element(54,     'Xe',   '')
Cs =    Element(55,     'Cs',   '')
Ba =    Element(56,     'Ba',   '')
La =    Element(57,     'La',   '')
Ce =    Element(58,     'Ce',   '')
Pr =    Element(59,     'Pr',   '')
Nd =    Element(60,     'Nd',   '')
Pm =    Element(61,     'Pm',   '')
Sm =    Element(62,     'Sm',   '')
Eu =    Element(63,     'Eu',   '')
Gd =    Element(64,     'Gd',   '')
Tb =    Element(65,     'Tb',   '')
Dy =    Element(66,     'Dy',   '')
Ho =    Element(67,     'Ho',   '')
Er =    Element(68,     'Er',   '')
Tm =    Element(69,     'Tm',   '')
Yb =    Element(70,     'Yb',   '')
Lu =    Element(71,     'Lu',   '')
Hf =    Element(72,     'Hf',   '')
Ta =    Element(73,     'Ta',   '')
W =     Element(74,     'W',    '')
Re =    Element(75,     'Re',   '')
Os =    Element(76,     'Os',   '')
Ir =    Element(77,     'Ir',   '')
Pt =    Element(78,     'Pt',   '')
Au =    Element(79,     'Au',   '')
Hg =    Element(80,     'Hg',   '')
Tl =    Element(81,     'Tl',   '')
Pb =    Element(82,     'Pb',   '')
Bi =    Element(83,     'Bi',   '')
Po =    Element(84,     'Po',   '')
At =    Element(85,     'At',   '')
Rn =    Element(86,     'Rn',   '')
Fr =    Element(87,     'Fr',   '')
Ra =    Element(88,     'Ra',   '')
Ac =    Element(89,     'Ac',   '')
Th =    Element(90,     'Th',   '')
Pa =    Element(91,     'Pa',   '')
U =     Element(92,     'U',    '')
Np =    Element(93,     'Np',   '')
Pu =    Element(94,     'Pu',   '')
Am =    Element(95,     'Am',   '')
Cm =    Element(96,     'Cm',   '')
Bk =    Element(97,     'Bk',   '')
Cf =    Element(98,     'Cf',   '')
Es =    Element(99,     'Es',   '')
Fm =    Element(100,    'Fm',   '')
Md =    Element(101,    'Md',   '')
No =    Element(102,    'No',   '')
Lr =    Element(103,    'Lr',   '')
Rf =    Element(104,    'Rf',   '')
Db =    Element(105,    'Db',   '')
Sg =    Element(106,    'Sg',   '')
Bh =    Element(107,    'Bh',   '')
Hs =    Element(108,    'Hs',   '')
Mt =    Element(109,    'Mt',   '')
Ds =    Element(110,    'Ds',   '')
Rg =    Element(111,    'Rg',   '')
Cn =    Element(112,    'Cn',   '')
Nh =    Element(113,    'Nh',   '')
Fl =    Element(114,    'Fl',   '')
Mc =    Element(115,    'Mc',   '')
Lv =    Element(116,    'Lv',   '')
Ts =    Element(117,    'Ts',   '')
Og =    Element(118,    'Og',   '')

elements: list[Element] = [
    n,
    H,              He,
    Li, Be, B, C, N, O, F, Ne, Na, Mg, Al, Si, P, S, Cl, Ar, K, Ca, Sc, Ti, V, Cr, Mn, Fe,
    Co, Ni, Cu, Zn, Ga, Ge, As, Se, Br, Kr, Rb, Sr, Y, Zr, Nb, Mo, Tc, Ru, Rh, Pd, Ag, Cd, In,
    Sn, Sb, Te, I, Xe, Cs, Ba, La, Ce, Pr, Nd, Pm, Sm, Eu, Gd, Tb, Dy, Ho, Er, Tm, Yb, Lu, Hf,
    Ta, W, Re, Os, Ir, Pt, Au, Hg, Tl, Pb, Bi, Po, At, Rn, Fr, Ra, Ac, Th, Pa, U, Np, Pu, Am, Cm,
    Bk, Cf, Es, Fm, Md, No, Lr, Rf, Db, Sg, Bh, Hs, Mt, Ds, Rg, Cn, Nh, Fl, Mc, Lv, Ts, Og,
]

symbol_element: dict[str, Element] = {e.symbol: e for e in elements}
atomicnum_element: dict[int, Element] = {e.num: e for e in elements}



from __future__ import annotations

from chemrecon.schema.direction import Direction
from chemrecon.entrygraph.scoring import Scorer
from chemrecon.schema.entry_types.aam import AAM
from chemrecon.schema.entry_types.enzyme import Enzyme
from chemrecon.schema.entry_types.reaction import Reaction
from chemrecon.schema.procedural_relation_entrygraph import ProceduralRelationEG
from chemrecon.schema.relation_types_composed.reaction_has_aam_relation import ReactionHasAAM
from chemrecon.schema.relation_types_source.reaction_has_enzyme_relation import ReactionHasEnzyme
from chemrecon.schema.relation_types_source.reaction_reference_relation import ReactionReference
from chemrecon.entrygraph.explorationprotocol import ExplorationProtocol

ProtocolSelectAAM = ExplorationProtocol(
    relation_types = {
        (ReactionReference, Direction.SYMMETRIC),
        (ReactionHasEnzyme, Direction.BOTH),
        (ReactionHasAAM, Direction.FORWARDS)
    },
    # TODO filter (filter out name entries)
)
eg_scorer_reaction_select_aam = Scorer(
    score_entry_type = AAM,
    # TODO more advanced weighing
)

class ReactionSelectAAM(ProceduralRelationEG[Reaction, AAM]):
    """ Procedural relation which computes the most related AAMs for a given reaction using entry graph scoring.
    """

    # Database
    _table_name = 'ReactionSelectAAM'
    source_entrytype = Reaction
    target_entrytype = AAM

    # For EG
    protocol = ProtocolSelectAAM
    entrygraph_scorer = eg_scorer_reaction_select_aam



from typing import Optional

from chemrecon.schema import Column, Relation, Reaction, Compound
from chemrecon.schema.db_object import InverseRelation
class ReactionInvolvesCompound(Relation[Reaction, Compound]):
    """ Each reaction is connected by this relation to the compounds which take part in the reaction, annotated
        with the stoichiometric coefficient.
    """
    # Attributes
    n: int  #: The stoichiometric coefficient of the compound in the reaction

    # Database
    entrytype_name = 'Reaction involves Compound'
    _table_name = 'ReactionInvolvesCompound'
    symmetric = False
    source_entrytype = Reaction
    target_entrytype = Compound
    _attribute_columns = [
        Column('n', int)
    ]
    _index = [0]

    def __init__(
        self,
        n: int,
        recon_id_1: Optional[int] = None,
        recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.n = n

    def _vis_str(self) -> str:
        return f'has compound ({self.n})'

class CompoundParticipatesInReaction(InverseRelation[Compound, Reaction]):
    """ Each reaction is connected by this relation to the compounds which take part in the reaction, annotated
        with the stoichiometric coefficient.
    """
    # Attributes
    n: int  #: The stoichiometric coefficient of the compound in the reaction

    # Inverse
    inverse_main_relation = ReactionInvolvesCompound

    # Database
    entrytype_name = 'Compound Participates in Reaction'
    _table_name = 'CompoundParticipatesInReaction'
    symmetric = False
    source_entrytype = Compound
    target_entrytype = Reaction
    _attribute_columns = [
        Column('n', int)
    ]
    _index = [0]

    def __init__(
        self,
        n: int,
        recon_id_1: Optional[int] = None,
        recon_id_2: Optional[int] = None
    ):
        super().__init__(recon_id_1, recon_id_2)
        self.n = n

    def _vis_str(self) -> str:
        return f'in reaction ({self.n})'

# Set inverse
ReactionInvolvesCompound.has_inverse = CompoundParticipatesInReaction



from collections.abc import Callable
from typing import Any, cast, ParamSpec, TypeVar

def copy_signature[**P, T](_origin: Callable[P, T]) -> Callable[[Callable[..., Any]], Callable[P, T]]:
    """ Decorates a function to inform type checker that it takes the same arguments as another function.
    """
    def decorator(target_function: Callable[..., Any]) -> Callable[P, T]:
        return cast(Callable[P, T], target_function)

    return decorator


""" Code to handle S_GML from Juri's M-CSA tool
"""
from __future__ import annotations

import re
from typing import Optional

import matplotlib.pyplot as plt
import networkx as nx
from rdkit import Chem as rdk

from chemrecon.chem.mol import Mol

# Regex patterns
rule_pattern = re.compile(
    r'^rule \[(?:ruleID \".*\")?(?:left \[(.*?)\])?context \[(.*?)\](?:right \[(.*?)\])?\]$'
)
label_sub = re.compile(
    r'id (\d+) label \"([A-Z][a-z]?\d?\+?\-?|Amino\(..?, ..., \d+, [\w|*]\)|Alias\(.,.\)|\*)\"'
)
label_rep = r'id \g<1> label "\g<1>: \g<2>"'


def sanitise_gml(gml: str) -> str:
    r = f'{gml}'
    r.replace('\n', '').replace('\t', '')
    return r


class GMLException(Exception):
    pass


class GMLEmptyException(Exception):
    pass


class GMLEntry:
    entry_index: int
    proposals: list[list[GMLStep]]
    map: dict[int, int]  # For proposal 0

    def __init__(self, index: int):
        self.entry_index = index
        self.proposals = []

    def add_step(self, proposal_index: int, step: GMLStep):
        # Extend proposals list
        while len(self.proposals) < proposal_index:
            self.proposals.append([])

        # Add step
        self.proposals[proposal_index - 1].append(step)

    def set_map(self, aam: dict[int, int]):
        self.map = aam
        sanity_check(self.left_graph(), self.right_graph(), self.map)

    # Getters always get proposal 0 (best star rating in M-CSA)
    def left_graph(self) -> GMLGraph:
        return self.proposals[0][0].left_graph()

    def right_graph(self) -> GMLGraph:
        return self.proposals[0][-1].right_graph()

    def atom_atom_map(self) -> dict[int, int]:
        return self.map

class GMLStep:
    step_index: int
    gml: str
    lhs: GMLGraph
    rhs: GMLGraph

    # Pure gml
    left_gml: str
    context_gml: str
    right_gml: str

    def __init__(self, step_index: int, gml: str):
        self.step_index = step_index
        self.gml = str(gml).replace('\n', '').replace('\t', '')

        # Make S_GML for educt and product sides
        match = re.match(rule_pattern, self.gml)
        self.left_gml = match.group(1)
        self.context_gml = match.group(2)
        self.right_gml = match.group(3)

        lhs = (f'graph [\n {self.left_gml} {self.context_gml} \n]'
               if self.left_gml else f'graph [\n{self.context_gml}\n]')
        rhs = (f'graph [\n {self.context_gml} {self.right_gml} \n]'
               if self.right_gml else f'graph [\n{self.context_gml}\n]')

        self.lhs = GMLGraph(
            gml = lhs
        )
        self.rhs = GMLGraph(
            gml = rhs
        )

    def left_graph(self) -> GMLGraph:
        return self.lhs

    def right_graph(self) -> GMLGraph:
        return self.rhs


class GMLGraph:
    graph: nx.Graph
    gml: str
    component_subsets: list[frozenset[int]]
    component_subgraphs: list[nx.Graph]

    def __init__(self, gml: str, graph: nx.Graph = None):
        self.gml = sanitise_gml(gml)

        # TODO raise exception of called with empty graph

        if graph:
            self.graph = graph
        else:
            # Make graph from gml
            self.graph = nx.parse_gml(
                gml,
                label = 'id'
            )
        self.component_subgraphs = []
        self.component_subsets = []

    def get_component_subsets(self) -> list[frozenset[int]]:
        if not self.component_subsets:
            self.component_subsets = list(
                frozenset(s)
                for s in nx.connected_components(self.graph)
            )
        return self.component_subsets

    def get_components_subgraphs(self) -> list[nx.Graph]:
        if self.component_subgraphs:
            return self.component_subgraphs
        else:
            if not self.component_subsets:
                # Make subsets
                self.get_component_subsets()
            # Set the subgraphs
            self.component_subgraphs = [
                nx.subgraph(self.graph, compset) for compset in self.component_subsets
            ]
            return self.component_subgraphs

    def remove_hydrogens(self):
        g_ = self.graph.copy()
        to_remove: list = list()
        for n in g_.nodes():
            if g_.nodes[n]['label'] == 'H' and g_.degree[n] > 0:
                to_remove.append(n)
        g_.remove_nodes_from(to_remove)
        self.graph = g_

    def get_subgraph(self, indices: set[int]) -> nx.Graph:
        return self.graph.subgraph(indices)

    def draw(self):
        # Draw with networkx and matplotlib - need to display atom numbers on label!
        plt.figure(1, figsize = (8, 8), dpi = 240)
        pos = nx.spring_layout(self.graph)

        # Graph structure and node labels
        nx.draw(
            self.graph, pos,
            with_labels = True,
            node_size = 50,
            font_size = 6,
            labels = {
                k: f"{k}: {n['label']}" for k, n in self.graph.nodes.items()
            }
        )

        # Edge labels
        edict = dict()
        for (u, v, d) in self.graph.edges.data('label'):
            edict[(u, v)] = d

        nx.draw_networkx_edge_labels(
            self.graph, pos, edict, font_color = 'blue', font_size = 6
        )

        # Display graph
        plt.show()
        # TODO


# ChemGraph generation
def chemgraph_from_mol(mol: Mol) -> GMLGraph:
    # Based on an RDK mol, create a ChemGraph object.
    struct_graph = nx.Graph()
    for atom in mol.mol.GetAtoms():
        # Make the label based on element and charge
        fcharge = atom.GetFormalCharge()
        symbol = atom.GetSymbol()

        # Do not include hydrogen
        if symbol == 'H' and len(atom.GetNeighbors()) > 0:
            continue

        # Make label
        label: str = 'ERR'
        match fcharge:
            case 0:
                label = symbol
            case -1:
                label = f'{symbol}-'
            case 1:
                label = f'{symbol}+'
            case _ if fcharge < 0:
                label = f'{symbol}{fcharge}-'
            case _ if fcharge >= 0:
                label = f'{symbol}{fcharge}+'

        struct_graph.add_node(
            atom.GetIdx(),
            label = label
        )

    for bond in mol.mol.GetBonds():
        label: str = 'ERR'
        match bond.GetBondType():
            case rdk.BondType.SINGLE:
                label = '-'
            case rdk.BondType.DOUBLE:
                label = '='
            case rdk.BondType.TRIPLE:
                label = '#'
            case rdk.BondType.AROMATIC:
                label = ':'
            case _:
                raise RuntimeError(f'Unimplemented bond-type {bond.GetBondType()}')

        struct_graph.add_edge(
            bond.GetBeginAtomIdx(),
            bond.GetEndAtomIdx(),
            label = label
        )

    # Finalize and add the graph to the list of graphs to match
    cgraph = GMLGraph(
        gml = '',
        graph = struct_graph
    )
    return cgraph


# Entry serialisation
# ----------------------------------------------------------------------------------------------------------------------
def graph_to_gml(graph: nx.Graph, relabel_map: Optional[dict[int, int]] = None) -> str:
    """ Only returns 'inner' gml.
    """
    if relabel_map is None:
        relabel_map = {x: x for x in graph.nodes}

    s = '[\n'
    # s = 'graph [\n'
    # Nodes
    for n_index in graph.nodes:
        label = graph.nodes[n_index]['label']
        s += f'\tnode [ id {relabel_map[n_index]} label "{label}"]\n'

    # Edges
    for x, y in graph.edges:
        label = graph.edges[(x, y)]['label']
        s += f'\tedge [ source {relabel_map[x]} target {relabel_map[y]} label "{label}" ]\n'
    s += '\n]'
    return s

# Reaction graph sanitation and handling -------------------------------------------------------------------------------
def remove_residues(
    chemgraph: GMLGraph
) -> GMLGraph:
    """ Given a chemgraph representing either side of a reaction, return the same graph with identified
        catalytic enzyme residues removed.
    """
    g: nx.Graph = chemgraph.graph
    remove_nodes: set = set()  #

    components = list(nx.connected_components(g))
    for component in components:
        for v in component:
            l: str = g.nodes[v]['label']
            if l.startswith('Amino'):
                # Remove the entire component if it contains an amino acid
                remove_nodes = remove_nodes.union(component)
                break

    # Remove the nodes
    g_ = g.copy(as_view = False)
    g_.remove_nodes_from(remove_nodes)
    if g_.number_of_nodes() == 0:
        raise GMLEmptyException
    return GMLGraph(gml = '', graph = g_)


# Misc
# --------------------------------------------------------------------------------------------------------------
# Matching functions for graph similarity
def nmatch(n1, n2):
    try:
        return n1['label'] == n2['label']
    except KeyError:
        return 1


def ematch(e1, e2):
    try:
        return e1['label'] == e2['label']
    except KeyError:
        return 1

def sanity_check(l_graph: GMLGraph, r_graph: GMLGraph, map: dict[int, int]):
    for l_index, r_index in map.items():
        l_label: str = l_graph.graph.nodes[l_index]['label']
        r_label: str = r_graph.graph.nodes[r_index]['label']
        if l_label.removesuffix('-').removesuffix('+') != r_label.removesuffix('-').removesuffix('+'):
            assert False, 'Wrong match'



""" Methods for producing a ranking on an entry graph.
"""
from collections import OrderedDict
from typing import Callable, Optional, TYPE_CHECKING

import rustworkx

from chemrecon.entrygraph.entrygraph import (
    Edge, EntryGraph, SourceEdgeArtificial, SourceVertexArtificial, Vertex,
)
from chemrecon.schema import Entry, Relation

if TYPE_CHECKING:
    from chemrecon.schema.procedural_relation_entrygraph import ProceduralRelationEG


class Scorer[T_rank: Entry]:
    """ A scorer is a callable which takes an entrygraph and produces a ranking of the vertices according to
        the parameters of the scorer.

        The score of an entry is (informally) the probability that a random walk starting at one of the initial entries
        of the entry graph will terminate at that entry.
        The parameters of the random walk can be customized by specifying weights (probabilities) using a weight
        function on entries and relations, which alters the probability of choosing a given path.
        The default weight of all entries and relations is 1.
        For example, if you do not trust a particular source, edges and vertices from that source can have their
        weight reduced, making them count less in the scoring algorithm.

        A damping factor, `alpha` can be specified.
        With probability `1-alpha`, the random walk will choose to go to a random entry rather than continuing the walk.
        Furthermore, a _decay_factor_ can be specified such that entries further away from the initial entries are given
        lower scores.
        A decay factor of `0` disables this adjustment. The default is `0.2`.

        Scores are normalized such that the sum of scores is **1**, which allows comparing scores across entry graphs.

        Formally, the scores are computed using the PageRank algorithm (https://en.wikipedia.org/wiki/PageRank),
        starting from the initial vertices, and with dangling vertices pointing back to all initial vertices with
        equal probability.

    """
    score_entry_type: type[T_rank]

    entry_weight: Callable[[Entry], float]
    relation_weight: Callable[[Relation], float]

    # Algorithm parameters
    alpha: float
    decay_factor: float

    def __init__(
            self,
            score_entry_type: type[Entry],
            alpha: float = 0.85,
            decay_factor: float = 0.2,
            entry_weight: Optional[Callable[[Entry], float]] = None,
            relation_weight: Optional[Callable[[Relation], float]] = None,
    ):
        """ Specify a scorer.
        """
        self.alpha = alpha
        self.decay_factor = decay_factor

        self.score_entry_type = score_entry_type

        if entry_weight is not None:
            self.entry_weight = entry_weight
        else:
            self.entry_weight = lambda e: 1

        if relation_weight is not None:
            self.relation_weight = relation_weight
        else:
            self.relation_weight = lambda r: 1

    def __call__(self, entrygraph: EntryGraph) -> OrderedDict[T_rank, float]:
        """ Produces a ranking of the entries of the type `score_entry_type`.
            The result is an `OrderedDict`, with entries given in descending order of score.
        """
        from chemrecon.schema.procedural_relation_entrygraph import ProceduralRelationEG

        g = entrygraph.g.copy()

        # Add source vertex and connect to initial vertices
        source_vertex_index = g.add_node(SourceVertexArtificial())
        for init_v_index in entrygraph.initial_vertices:
            g.add_edge(source_vertex_index, init_v_index, SourceEdgeArtificial())

        # Get the weight for all edges
        edge_weight_dict: dict[Edge, float] = dict()
        for e_index in g.edge_indices():
            e = g.get_edge_data_by_index(e_index)
            source_index = g.get_edge_endpoints_by_index(e_index)[0]
            if isinstance(e, Edge):
                edge_weight_dict[e] = (
                        self.relation_weight(e.relation)
                        * self.entry_weight(g.get_node_data(source_index).entry)
                )
            else:
                edge_weight_dict[e] = 1

            # Modify edges by 'score' parameter of EG-based procedural relations
            if isinstance(e, Edge):
                if isinstance(e.relation, ProceduralRelationEG):
                    edge_weight_dict[e] *= e.relation.score

        for e_index in g.edge_indices():
            e = g.get_edge_data_by_index(e_index)

        # Produce the initial score (ranking all vertices, not normalized)
        init_scoring = rustworkx.pagerank(
            g,
            alpha = self.alpha,
            weight_fn = lambda e_: edge_weight_dict.get(e_, 1),
            dangling = {source_vertex_index: 1},  # Terminal nodes should loop back to source
            personalization = {source_vertex_index: 1}  # Random traversals go back to source
        )

        # Use only eligible entries (filter by relevant entry type)
        scoring: dict[Vertex, float] = dict()
        for v_idx, score in init_scoring.items():
            v = g.get_node_data(v_idx)
            if isinstance(v, Vertex) and isinstance(v.entry, self.score_entry_type):
                scoring[v] = score

        # Apply decay based on the generation
        if self.decay_factor != 0:
            for v, score in scoring.items():
                v: Vertex
                scoring[v] = score * ((1 - self.decay_factor)**v.generation)

        # Normalise s.t. sum(scoring.values()) == 1.0
        score_sum = sum(s for _, s in scoring.items())
        for v, score in scoring.items():
            scoring[v] = score / score_sum

        # Finalise
        return OrderedDict(
            (k.entry, v) for k, v in
            sorted(scoring.items(), key = lambda pair: pair[1], reverse = True)
        )



""" Implements database connection, cursor, etc. functionality.
"""
from __future__ import annotations

import psycopg as pg
import psycopg.sql as sql
from psycopg.types.enum import EnumInfo, register_enum

import chemrecon.schema as schema
from chemrecon.database.params import Params

# Initialisation functions
# ----------------------------------------------------------------------------------------------------------------------
def postgres_connect(
    params: Params,
    initialise_enums: bool = True
) -> pg.Connection:
    """ Connect to the database server.
        Connection string given relative to the working directory (chemrecon/).
        If find_directory is set, will automatically try to locate the chemrecon directory
    """
    # Load parameters
    print(f'[ChemRecon] Attempting to connect with string: {params.connection_string()}')
    try:
        conn: pg.Connection = pg.connect(
            params.connection_string()
        )
    except pg.OperationalError as e:
        print(f'[ChemRecon] Could not connect: {e}')
        raise ConnectionError from e

    print(f'[ChemRecon] Connection succesful.')

    # Set correct schemata to search
    schemata = [
        'chemrecon',
        'pg_catalog',
        'meta',
        'public'
    ]
    conn.execute(
        sql.SQL("SET {search_path} TO {schemata}").format(
            search_path = sql.Identifier('search_path'),
            schemata = sql.SQL(', ').join(
                [sql.Identifier(s) for s in schemata]
            )
        )
    )
    conn.commit()

    # Register enum mapping
    if initialise_enums:
        conn.commit()
        for enum in schema.enums.enum_register:
            #  enum_info = EnumInfo.fetch(conn, f'chemrecon.{enum.__name__}')
            enum_info = EnumInfo.fetch(conn, sql.Identifier(enum.__name__))
            if enum_info is None:
                raise RuntimeError(f'Enum info none for enum {enum}.')
            register_enum(enum_info, conn, enum)

    # Done
    conn.commit()
    return conn